# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Download file.
"""

import shutil
import tarfile
import zipfile
import six
import requests
import os
import sys
from tqdm import tqdm


def http_get(url, path, extract: bool = True):
    """
    Downloads a URL to a given path on disc
    """
    if os.path.dirname(path) != '':
        os.makedirs(os.path.dirname(path), exist_ok=True)

    req = requests.get(url, stream=True)
    if req.status_code != 200:
        print("Exception when trying to download {}. Response {}".format(url, req.status_code), file=sys.stderr)
        req.raise_for_status()
        return

    download_filepath = path + "_part"
    with open(download_filepath, "wb") as file_binary:
        content_length = req.headers.get('Content-Length')
        total = int(content_length) if content_length is not None else None
        progress = tqdm(unit="B", total=total, unit_scale=True)
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                progress.update(len(chunk))
                file_binary.write(chunk)

    os.rename(download_filepath, path)
    progress.close()

    if extract:
        data_dir = os.path.dirname(os.path.abspath(path))
        _extract_archive(path, data_dir, 'auto')



def _extract_archive(file_path, path='.', archive_format='auto'):
    """
    Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    :param file_path: path to the archive file
    :param path: path to extract the archive file
    :param archive_format: Archive format to try for extracting the file.
        Options are 'auto', 'tar', 'zip', and None.
        'tar' includes tar, tar.gz, and tar.bz files.
        The default 'auto' is ['tar', 'zip'].
        None or an empty list will return no matches found.

    :return: True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == 'auto':
        archive_format = ['tar', 'zip']
    if isinstance(archive_format, six.string_types):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type == 'tar':
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type == 'zip':
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError,
                        KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False



# -*- coding: utf-8 -*-
"""
refer: https://github.com/UKPLab/beir/blob/main/beir/datasets/data_loader.py
"""
from typing import List, Dict, Tuple

from loguru import logger


def mrr(qrels: Dict[str, Dict[str, int]],
        results: Dict[str, Dict[str, float]],
        k_values: List[int]) -> Tuple[Dict[str, float]]:
    mrr = {}

    for k in k_values:
        mrr[f"MRR@{k}"] = 0.0

    k_max, top_hits = max(k_values), {}
    logger.info("\n")

    for query_id, doc_scores in results.items():
        top_hits[query_id] = sorted(doc_scores.items(), key=lambda item: item[1], reverse=True)[0:k_max]

    for query_id in qrels:
        query_relevant_docs = set([doc_id for doc_id in qrels[query_id] if qrels[query_id][doc_id] > 0])
        for k in k_values:
            for rank, hit in enumerate(top_hits[query_id][0:k]):
                if hit[0] in query_relevant_docs:
                    mrr[f"MRR@{k}"] += 1.0 / (rank + 1)
                    break

    for k in k_values:
        mrr[f"MRR@{k}"] = round(mrr[f"MRR@{k}"] / len(qrels), 5)
        logger.info("MRR@{}: {:.4f}".format(k, mrr[f"MRR@{k}"]))

    return mrr


def recall_cap(qrels: Dict[str, Dict[str, int]],
               results: Dict[str, Dict[str, float]],
               k_values: List[int]) -> Tuple[Dict[str, float]]:
    capped_recall = {}

    for k in k_values:
        capped_recall[f"R_cap@{k}"] = 0.0

    k_max = max(k_values)
    logger.info("\n")

    for query_id, doc_scores in results.items():
        top_hits = sorted(doc_scores.items(), key=lambda item: item[1], reverse=True)[0:k_max]
        query_relevant_docs = [doc_id for doc_id in qrels[query_id] if qrels[query_id][doc_id] > 0]
        for k in k_values:
            retrieved_docs = [row[0] for row in top_hits[0:k] if qrels[query_id].get(row[0], 0) > 0]
            denominator = min(len(query_relevant_docs), k)
            capped_recall[f"R_cap@{k}"] += (len(retrieved_docs) / denominator)

    for k in k_values:
        capped_recall[f"R_cap@{k}"] = round(capped_recall[f"R_cap@{k}"] / len(results), 5)
        logger.info("R_cap@{}: {:.4f}".format(k, capped_recall[f"R_cap@{k}"]))

    return capped_recall


def hole(qrels: Dict[str, Dict[str, int]],
         results: Dict[str, Dict[str, float]],
         k_values: List[int]) -> Tuple[Dict[str, float]]:
    hole = {}

    for k in k_values:
        hole[f"Hole@{k}"] = 0.0

    annotated_corpus = set()
    for _, docs in qrels.items():
        for doc_id, score in docs.items():
            annotated_corpus.add(doc_id)

    k_max = max(k_values)
    logger.info("\n")

    for _, scores in results.items():
        top_hits = sorted(scores.items(), key=lambda item: item[1], reverse=True)[0:k_max]
        for k in k_values:
            hole_docs = [row[0] for row in top_hits[0:k] if row[0] not in annotated_corpus]
            hole[f"Hole@{k}"] += len(hole_docs) / k

    for k in k_values:
        hole[f"Hole@{k}"] = round(hole[f"Hole@{k}"] / len(results), 5)
        logger.info("Hole@{}: {:.4f}".format(k, hole[f"Hole@{k}"]))

    return hole


def top_k_accuracy(
        qrels: Dict[str, Dict[str, int]],
        results: Dict[str, Dict[str, float]],
        k_values: List[int]) -> Tuple[Dict[str, float]]:
    top_k_acc = {}

    for k in k_values:
        top_k_acc[f"Accuracy@{k}"] = 0.0

    k_max, top_hits = max(k_values), {}
    logger.info("\n")

    for query_id, doc_scores in results.items():
        top_hits[query_id] = [item[0] for item in
                              sorted(doc_scores.items(), key=lambda item: item[1], reverse=True)[0:k_max]]

    for query_id in qrels:
        query_relevant_docs = set([doc_id for doc_id in qrels[query_id] if qrels[query_id][doc_id] > 0])
        for k in k_values:
            for relevant_doc_id in query_relevant_docs:
                if relevant_doc_id in top_hits[query_id][0:k]:
                    top_k_acc[f"Accuracy@{k}"] += 1.0
                    break

    for k in k_values:
        top_k_acc[f"Accuracy@{k}"] = round(top_k_acc[f"Accuracy@{k}"] / len(qrels), 5)
        logger.info("Accuracy@{}: {:.4f}".format(k, top_k_acc[f"Accuracy@{k}"]))

    return top_k_acc


def evaluate(
        qrels: Dict[str, Dict[str, int]],
        results: Dict[str, Dict[str, float]],
        k_values: List[int] = (1, 3, 5, 10, 100)
) -> Tuple[Dict[str, float], Dict[str, float], Dict[str, float], Dict[str, float]]:
    try:
        import pytrec_eval
    except ImportError:
        raise ImportError("Please install pytrec_eval to use this function, eg. `pip install pytrec_eval`")

    ndcg = {}
    _map = {}
    recall = {}
    precision = {}

    for k in k_values:
        ndcg[f"NDCG@{k}"] = 0.0
        _map[f"MAP@{k}"] = 0.0
        recall[f"Recall@{k}"] = 0.0
        precision[f"P@{k}"] = 0.0

    map_string = "map_cut." + ",".join([str(k) for k in k_values])
    ndcg_string = "ndcg_cut." + ",".join([str(k) for k in k_values])
    recall_string = "recall." + ",".join([str(k) for k in k_values])
    precision_string = "P." + ",".join([str(k) for k in k_values])
    evaluator = pytrec_eval.RelevanceEvaluator(qrels, {map_string, ndcg_string, recall_string, precision_string})
    scores = evaluator.evaluate(results)

    for query_id in scores.keys():
        for k in k_values:
            ndcg[f"NDCG@{k}"] += scores[query_id]["ndcg_cut_" + str(k)]
            _map[f"MAP@{k}"] += scores[query_id]["map_cut_" + str(k)]
            recall[f"Recall@{k}"] += scores[query_id]["recall_" + str(k)]
            precision[f"P@{k}"] += scores[query_id]["P_" + str(k)]

    for k in k_values:
        ndcg[f"NDCG@{k}"] = round(ndcg[f"NDCG@{k}"] / len(scores), 5)
        _map[f"MAP@{k}"] = round(_map[f"MAP@{k}"] / len(scores), 5)
        recall[f"Recall@{k}"] = round(recall[f"Recall@{k}"] / len(scores), 5)
        precision[f"P@{k}"] = round(precision[f"P@{k}"] / len(scores), 5)

    for eval in [ndcg, _map, recall, precision]:
        logger.info("\n")
        for k in eval.keys():
            logger.info("{}: {:.4f}".format(k, eval[k]))

    return ndcg, _map, recall, precision


def evaluate_custom(
        qrels: Dict[str, Dict[str, int]],
        results: Dict[str, Dict[str, float]],
        k_values: List[int] = (1, 3, 5, 10, 100),
        metric: str = 'acc') -> Tuple[Dict[str, float]]:
    if metric.lower() in ["mrr", "mrr@k", "mrr_cut"]:
        return mrr(qrels, results, k_values)

    elif metric.lower() in ["recall_cap", "r_cap", "r_cap@k"]:
        return recall_cap(qrels, results, k_values)

    elif metric.lower() in ["hole", "hole@k"]:
        return hole(qrels, results, k_values)

    elif metric.lower() in ["acc", "top_k_acc", "accuracy", "accuracy@k", "top_k_accuracy"]:
        return top_k_accuracy(qrels, results, k_values)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: CLIP model for text and image embeddings
"""
import math
import queue
from typing import List, Union, Dict

import numpy as np
import torch
import torch.multiprocessing as mp
import torch.nn.functional
from PIL import Image
from loguru import logger
from torch import nn
from tqdm import trange
from transformers import ChineseCLIPProcessor, ChineseCLIPModel, CLIPProcessor, CLIPModel


class ClipModule(nn.Module):
    """
    CLIP model for text and image embeddings

    Args:
        model_name_or_path: str, default "OFA-Sys/chinese-clip-vit-base-patch16"
            chinese model url: https://huggingface.co/OFA-Sys/chinese-clip-vit-base-patch16
            english model url: https://huggingface.co/openai/clip-vit-base-patch32
        processor_name: str, default None
        device: str, default None
        is_chinese_model: bool, default None, if None, auto detect by model_name_or_path
    """

    def __init__(
            self,
            model_name_or_path: str = "OFA-Sys/chinese-clip-vit-base-patch16",
            processor_name: str = None,
            device: str = None,
            is_chinese_model: bool = None,
    ):
        super(ClipModule, self).__init__()
        if device is None:
            self.device = "cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu"
            logger.debug(f"Device: {self.device}")
        else:
            self.device = device
        self.model_name_or_path = model_name_or_path
        if processor_name is None:
            processor_name = model_name_or_path
        if is_chinese_model is None:
            is_chinese_model = 'chinese' in model_name_or_path
        self.is_chinese_model = is_chinese_model
        if is_chinese_model:
            self.model = ChineseCLIPModel.from_pretrained(model_name_or_path)
            self.processor = ChineseCLIPProcessor.from_pretrained(processor_name)
        else:
            self.model = CLIPModel.from_pretrained(model_name_or_path)
            self.processor = CLIPProcessor.from_pretrained(processor_name)
        self.model.to(self.device)

    def __str__(self):
        return f"model_name_or_path: {self.model_name_or_path} ClipModule({self.model})"

    def forward(self, features):
        image_embeds = []
        text_embeds = []

        if 'pixel_values' in features:
            vision_outputs = self.model.vision_model(pixel_values=features['pixel_values'])
            image_embeds = self.model.visual_projection(vision_outputs[1])

        if 'input_ids' in features:
            text_outputs = self.model.text_model(
                input_ids=features.get('input_ids'),
                attention_mask=features.get('attention_mask', None),
                position_ids=features.get('position_ids', None),
                output_attentions=features.get('output_attentions', None),
                output_hidden_states=features.get('output_hidden_states', None),
            )
            if self.is_chinese_model:
                # refer chinese clip: https://github.com/huggingface/transformers/blob/main/src/transformers/models/chinese_clip/modeling_chinese_clip.py#L1431
                pooled_output = text_outputs[0][:, 0, :]
            else:
                pooled_output = text_outputs[1]
            text_embeds = self.model.text_projection(pooled_output)

        sentence_embedding = []
        image_features = iter(image_embeds)
        text_features = iter(text_embeds)

        for idx, input_type in enumerate(features['image_text_info']):
            if input_type == 0:
                sentence_embedding.append(next(image_features))
            else:
                sentence_embedding.append(next(text_features))

        features['embedding'] = torch.stack(sentence_embedding).float()

        return features

    def tokenize(self, texts):
        images = []
        texts_values = []
        image_text_info = []

        for idx, data in enumerate(texts):
            if isinstance(data, (Image.Image, np.ndarray)):  # An Image
                images.append(data)
                image_text_info.append(0)
            else:  # A text
                texts_values.append(data)
                image_text_info.append(1)

        if len(texts_values) == 0:
            texts_values = None
        if len(images) == 0:
            images = None

        inputs = self.processor(text=texts_values, images=images, return_tensors="pt", padding=True)
        inputs['image_text_info'] = image_text_info
        return inputs

    def save(self, output_path: str):
        self.model.save_pretrained(output_path)
        self.processor.save_pretrained(output_path)

    @staticmethod
    def load(input_path: str):
        return ClipModule(model_name_or_path=input_path)

    def _text_length(self, text):
        """
        Help function to get the length for the input text. Text can be either
        a list of ints (which means a single text as input), or a tuple of list of ints
        (representing several text inputs to the model).
        """

        if isinstance(text, dict):  # {key: value} case
            return len(next(iter(text.values())))
        elif not hasattr(text, '__len__'):  # Object has no len() method
            return 1
        elif len(text) == 0 or isinstance(text[0], int):  # Empty string or list of ints
            return len(text)
        else:
            return sum([len(t) for t in text])  # Sum of length of individual strings

    @staticmethod
    def batch_to_device(batch, device):
        """
        send a pytorch batch to a device (CPU/GPU)
        """
        for key in batch:
            if isinstance(batch[key], torch.Tensor):
                batch[key] = batch[key].to(device)
        return batch

    def encode(
            self,
            sentences: Union[str, List[str], Image.Image, List[Image.Image]],
            batch_size: int = 32,
            show_progress_bar: bool = False,
            convert_to_numpy: bool = True,
            convert_to_tensor: bool = False,
            normalize_embeddings: bool = False,
            device: str = None,
    ):
        """
        Computes sentence and images embeddings

        :param sentences: list of sentences or list of Image.Image
        :param batch_size: the batch size used for the computation
        :param show_progress_bar: Output a progress bar when encode sentences
        :param convert_to_numpy: If true, the output is a list of numpy vectors. Else, it is a list of pytorch tensors.
        :param convert_to_tensor: If true, the output is a stacked tensor. Else, it is a list of pytorch tensors.
        :param normalize_embeddings: If set to true, returned vectors will have length 1.
            In that case, the faster dot-product (util.dot_score) instead of cosine similarity can be used.
        :param device: Which device to use for the computation

        :return:
            By default, a list of tensors is returned. If convert_to_tensor, a stacked tensor is returned.
            If convert_to_numpy, a numpy matrix is returned.
        """
        if device is None:
            device = self.device
        self.model.eval()
        input_was_string = False
        if isinstance(sentences, str) or not hasattr(sentences, '__len__'):
            sentences = [sentences]
            input_was_string = True
        self.model.to(device)
        if convert_to_tensor:
            convert_to_numpy = False
        all_embeddings = []
        length_sorted_idx = np.argsort([-self._text_length(sent) for sent in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index:start_index + batch_size]
            features = self.tokenize(sentences_batch)
            features = self.batch_to_device(features, device)

            with torch.no_grad():
                out_features = self.forward(features)
                embeddings = out_features['embedding']
                embeddings = embeddings.detach()
                if normalize_embeddings:
                    embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
                # fixes for #522 and #487 to avoid oom problems on gpu with large datasets
                if convert_to_numpy:
                    embeddings = embeddings.cpu()
                all_embeddings.extend(embeddings)

        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]
        if convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])
        else:
            all_embeddings = torch.stack(all_embeddings)
        if input_was_string:
            all_embeddings = all_embeddings[0]

        return all_embeddings

    def start_multi_process_pool(self, target_devices: List[str] = None):
        """
        Starts multi processes to process the encoding with several, independent processes.
        This method is recommended if you want to encode on multiple GPUs. It is advised
        to start only one process per GPU. This method works together with encode_multi_process

        :param target_devices: PyTorch target devices,
            list, e.g. ['cuda:0', 'cuda:1'] If None, all available CUDA devices will be used
        :return: Returns a dict with the target processes, an input queue and output queue.
        """
        if target_devices is None:
            if torch.cuda.is_available():
                target_devices = ['cuda:{}'.format(i) for i in range(torch.cuda.device_count())]
            else:
                logger.info("CUDA is not available. Start 4 CPU worker")
                target_devices = ['cpu'] * 4

        logger.info("Start multi-process pool on devices: {}".format(', '.join(map(str, target_devices))))

        ctx = mp.get_context('spawn')
        input_queue = ctx.Queue()
        output_queue = ctx.Queue()
        processes = []

        for cuda_id in target_devices:
            p = ctx.Process(
                target=ClipModule._encode_multi_process_worker,
                args=(cuda_id, self, input_queue, output_queue),
                daemon=True
            )
            p.start()
            processes.append(p)

        return {'input': input_queue, 'output': output_queue, 'processes': processes}

    @staticmethod
    def stop_multi_process_pool(pool):
        """
        Stops all processes started with start_multi_process_pool
        """
        for p in pool['processes']:
            p.terminate()

        for p in pool['processes']:
            p.join()
            p.close()

        pool['input'].close()
        pool['output'].close()

    def encode_multi_process(
            self,
            sentences: Union[List[str], List[Image.Image]],
            pool: Dict[str, object],
            batch_size: int = 32,
            normalize_embeddings: bool = False,
            chunk_size: int = None
    ):
        """
        This method allows to run encode() on multiple GPUs. The sentences are chunked into smaller packages
        and sent to individual processes, which encode these on the different GPUs. This method is only suitable
        for encoding large sets of sentences

        :param sentences: List of sentences, or list of images
        :param pool: A pool of workers started with start_multi_process_pool
        :param batch_size: Encode sentences with batch size
        :param normalize_embeddings: bool, Whether to normalize embeddings before returning them
        :param chunk_size: Sentences are chunked and sent to the individual processes. If none, it is a sensible size.
        :return: Numpy matrix with all embeddings
        """

        if chunk_size is None:
            chunk_size = min(math.ceil(len(sentences) / len(pool["processes"]) / 10), 5000)

        logger.debug(f"Chunk data into {math.ceil(len(sentences) / chunk_size)} packages of size {chunk_size}")

        input_queue = pool['input']
        last_chunk_id = 0
        chunk = []

        for sentence in sentences:
            chunk.append(sentence)
            if len(chunk) >= chunk_size:
                input_queue.put([last_chunk_id, batch_size, chunk, normalize_embeddings])
                last_chunk_id += 1
                chunk = []

        if len(chunk) > 0:
            input_queue.put([last_chunk_id, batch_size, chunk, normalize_embeddings])
            last_chunk_id += 1

        output_queue = pool['output']
        results_list = sorted([output_queue.get() for _ in range(last_chunk_id)], key=lambda x: x[0])
        embeddings = np.concatenate([result[1] for result in results_list])
        return embeddings

    @staticmethod
    def _encode_multi_process_worker(target_device: str, model, input_queue, results_queue):
        """
        Internal working process to encode sentences in multi processes setup
        """
        while True:
            try:
                id, batch_size, sentences, normalize_embeddings = input_queue.get()
                embeddings = model.encode(
                    sentences,
                    device=target_device,
                    show_progress_bar=False,
                    convert_to_numpy=True,
                    batch_size=batch_size,
                    normalize_embeddings=normalize_embeddings,
                )
                results_queue.put([id, embeddings])
            except queue.Empty:
                break



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: cli entry point
"""
import sys

import fire

sys.path.append('..')
from similarities.faiss_bert_similarity import bert_embedding, bert_index, bert_filter, bert_server
from similarities.faiss_clip_similarity import clip_embedding, clip_index, clip_filter, clip_server


def main():
    """Main entry point"""

    fire.Fire(
        {
            "bert_embedding": bert_embedding,
            "bert_index": bert_index,
            "bert_filter": bert_filter,
            "bert_server": bert_server,
            "clip_embedding": clip_embedding,
            "clip_index": clip_index,
            "clip_filter": clip_filter,
            "clip_server": clip_server,
        }
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
from difflib import SequenceMatcher

import numpy as np
from similarities.utils.util import cos_sim

zero_bit = 1e-9


def try_divide(x, y, val=0.0):
    """
    try to divide two numbers
    """
    if y != 0.0:
        val = float(x) / y
    return val


def cosine_distance(v1, v2):
    """
    Compute the cosine distance between two vectors.
    return cos score
    """
    return cos_sim(v1, v2)


def hamming_distance(seq1, seq2, normalize=False):
    """Compute the Hamming distance between the two sequences `seq1` and `seq2`.
    The Hamming distance is the number of differing items in two ordered
    sequences of the same length. If the sequences submitted do not have the
    same length, an error will be raised.

    If `normalized` is `False`, the return value will be an integer
    between 0 and the length of the sequences provided, edge values included;
    otherwise, it will be a float between 0 and 1 included, where 0 means
    equal, and 1 totally different. Normalized hamming distance is computed as:

        0.0                         if len(seq1) == 0
        hamming_dist / len(seq1)    otherwise
    """
    L = len(seq1)
    if L != len(seq2):
        raise ValueError("expected two strings of the same length")
    if L == 0:
        return 0.0 if normalize else 0  # equal
    dist = sum(c1 != c2 for c1, c2 in zip(seq1, seq2))
    if normalize:
        return dist / float(L)
    return dist


def euclidean_distance(v1, v2, normalize=False):  # 欧氏距离
    score = np.sqrt(np.sum(np.square(v1 - v2)))
    if normalize:
        score = 1.0 / (1.0 + score)
    return score


def manhattan_distance(v1, v2):  # 曼哈顿距离
    return np.sum(np.abs(v1 - v2))


def chebyshev_distance(v1, v2):  # 切比雪夫距离
    return np.max(np.abs(v1 - v2))


def minkowski_distance(v1, v2):  # 闵可夫斯基距离
    return np.sqrt(np.sum(np.square(v1 - v2)))


def euclidean_distance_standardized(v1, v2):  # 标准化欧氏距离
    v1_v2 = np.vstack([v1, v2])
    sk_v1_v2 = np.var(v1_v2, axis=0, ddof=1)
    return np.sqrt(((v1 - v2) ** 2 / (sk_v1_v2 + zero_bit * np.ones_like(sk_v1_v2))).sum())


def edit_distance(str1, str2):
    try:
        # very fast
        # http://stackoverflow.com/questions/14260126/how-python-levenshtein-ratio-is-computed
        import Levenshtein
        d = Levenshtein.distance(str1, str2) / float(max(len(str1), len(str2)))
    except:
        # https://docs.python.org/2/library/difflib.html
        d = 1.0 - SequenceMatcher(lambda x: x == " ", str1, str2).ratio()
    return d


def pearson_correlation_distance(v1, v2):  # 皮尔逊相关系数（Pearson correlation）
    v1_v2 = np.vstack([v1, v2])
    return np.corrcoef(v1_v2)[0][1]


def jaccard_similarity_coefficient_distance(v1, v2):  # 杰卡德相似系数(Jaccard similarity coefficient)
    # 公式求解
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    up = np.double(np.bitwise_and((v1 != v2), np.bitwise_or(v1 != 0, v2 != 0)).sum())
    down = np.double(np.bitwise_or(v1 != 0, v2 != 0).sum() + zero_bit)
    return try_divide(up, down)


def is_str_match(str1, str2, threshold=1.0):
    assert 0.0 <= threshold <= 1.0, "Wrong threshold."
    if float(threshold) == 1.0:
        return str1 == str2
    else:
        return (1.0 - edit_distance(str1, str2)) >= threshold


def longest_match_size(str1, str2):
    """最长公共子串长度"""
    sq = SequenceMatcher(None, str1, str2)
    match = sq.find_longest_match(0, len(str1), 0, len(str2))
    return match.size


def longest_match_ratio(str1, str2):
    """最长公共子串占比"""
    sq = SequenceMatcher(None, str1, str2)
    match = sq.find_longest_match(0, len(str1), 0, len(str2))
    return try_divide(match.size, min(len(str1), len(str2)))


def jaccard_coef(A, B):
    if not isinstance(A, set):
        A = set(A)
    if not isinstance(B, set):
        B = set(B)
    return try_divide(float(len(A.intersection(B))), len(A.union(B)))


def num_of_common_sub_str(str1, str2):
    """
    求两个字符串的最长公共子串，同longest_match_size
    思想：建立一个二维数组，保存连续位相同与否的状态
    """
    lstr1 = len(str1)
    lstr2 = len(str2)
    record = [[0 for i in range(lstr2 + 1)] for j in range(lstr1 + 1)]  # 多一位
    max_num = 0  # 最长匹配长度

    for i in range(lstr1):
        for j in range(lstr2):
            if str1[i] == str2[j]:
                # 相同则累加
                record[i + 1][j + 1] = record[i][j] + 1
                if record[i + 1][j + 1] > max_num:
                    # 获取最大匹配长度
                    max_num = record[i + 1][j + 1]
    return max_num


def string_hash(source):
    if source == "":
        return 0
    else:
        x = ord(source[0]) << 7
        m = 1000003
        mask = 2 ** 128 - 1
        for c in source:
            x = ((x * m) ^ ord(c)) & mask
        x ^= len(source)
        if x == -1:
            x = -2
        x = bin(x).replace('0b', '').zfill(64)[-64:]

        return str(x)


def max_min_normalize(x):
    """
    最大最小值归一化
    :param x:
    :return:
    """
    return [(float(i) - min(x)) / float(max(x) - min(x) + zero_bit) for i in x]


def z_score(x, axis=0):
    """
    z_score标准化
    :param x: array, numpy
    :param axis: int, 0
    :return: np.array, numpy
    """
    x = np.array(x).astype(float)
    xr = np.rollaxis(x, axis=axis)
    xr -= np.mean(x, axis=axis)
    xr /= np.std(x, axis=axis)
    return x


if __name__ == '__main__':
    vec1_test = np.array([1.0, 38.0, 17.0, 32.0])
    vec2_test = np.array([5.0, 6.0, 8.0, 9.0])

    print(euclidean_distance(vec1_test, vec2_test))
    print(cosine_distance(vec1_test, vec2_test))
    print(manhattan_distance(vec1_test, vec2_test))

    str1_test = "你到底是谁?"
    str2_test = "没想到我是谁，是真样子"
    print('strs:', str1_test, ' vs ', str2_test)
    print('edit_dist', edit_distance(str1_test, str2_test))
    print('edit_sim:', 1 - edit_distance(str1_test, str2_test))

    str1_test = "private Thread currentThread;"
    str2_test = "private volatile Thread currentThread;"
    print('strs:', str1_test, ' vs ', str2_test)
    print('edit_dist', edit_distance(str1_test, str2_test))
    print('edit_sim:', 1 - edit_distance(str1_test, str2_test))

    print(num_of_common_sub_str(str1_test, str2_test))
    print(max_min_normalize(vec1_test))  # 归一化（0-1）
    print(z_score(vec1_test))  # 标准化（0附近，正负）

    str1 = '刘若英是演员和歌手'
    str2 = '刘若英是演员吗？'
    print(f"{str1} vs {str2} common sub str: {num_of_common_sub_str(str1, str2)}")
    print(f"{str1} vs {str2} longest match size: {longest_match_size(str1, str2)}")
    print(f"{str1} vs {str2} longest match ratio: {longest_match_ratio(str1, str2)}")



# -*- coding: utf-8 -*-
"""
refer: https://github.com/UKPLab/beir/blob/main/beir/datasets/data_loader.py
"""
import csv
import json
import os
from typing import Dict, Tuple

from loguru import logger
from tqdm.autonotebook import tqdm


class SearchDataLoader:

    def __init__(self, data_folder: str = None, prefix: str = None, corpus_file: str = "corpus.jsonl",
                 query_file: str = "queries.jsonl",
                 qrels_folder: str = "qrels", qrels_file: str = ""):
        self.corpus = {}
        self.queries = {}
        self.qrels = {}

        if prefix:
            query_file = prefix + "-" + query_file
            qrels_folder = prefix + "-" + qrels_folder

        self.corpus_file = os.path.join(data_folder, corpus_file) if data_folder else corpus_file
        self.query_file = os.path.join(data_folder, query_file) if data_folder else query_file
        self.qrels_folder = os.path.join(data_folder, qrels_folder) if data_folder else None
        self.qrels_file = qrels_file

    @staticmethod
    def check(fIn: str, ext: str):
        if not os.path.exists(fIn):
            raise ValueError("File {} not present! Please provide accurate file.".format(fIn))

        if not fIn.endswith(ext):
            raise ValueError("File {} must be present with extension {}".format(fIn, ext))

    def load_custom(self) -> Tuple[Dict[str, Dict[str, str]], Dict[str, str], Dict[str, Dict[str, int]]]:
        self.check(fIn=self.corpus_file, ext="jsonl")
        self.check(fIn=self.query_file, ext="jsonl")
        self.check(fIn=self.qrels_file, ext="tsv")

        if not len(self.corpus):
            logger.info("Loading Corpus...")
            self._load_corpus()
            logger.info(f"Loaded {len(self.corpus)} Documents.")
            logger.info(f"Doc Example: {list(self.corpus.values())[0]}")

        if not len(self.queries):
            logger.info("Loading Queries...")
            self._load_queries()

        if os.path.exists(self.qrels_file):
            self._load_qrels()
            self.queries = {qid: self.queries[qid] for qid in self.qrels}
            logger.info("Loaded {len(self.queries)} Queries.")
            logger.info("Query Example: {list(self.queries.values())[0]}")

        return self.corpus, self.queries, self.qrels

    def load(self, split="test") -> Tuple[Dict[str, Dict[str, str]], Dict[str, str], Dict[str, Dict[str, int]]]:
        self.qrels_file = os.path.join(self.qrels_folder, split + ".tsv")
        self.check(fIn=self.corpus_file, ext="jsonl")
        self.check(fIn=self.query_file, ext="jsonl")
        self.check(fIn=self.qrels_file, ext="tsv")

        if not len(self.corpus):
            logger.info("Loading Corpus...")
            self._load_corpus()
            logger.info(f"Loaded {len(self.corpus)} {split.upper()} Documents.")
            logger.info(f"Doc Example: {list(self.corpus.values())[0]}")

        if not len(self.queries):
            logger.info("Loading Queries...")
            self._load_queries()

        if os.path.exists(self.qrels_file):
            self._load_qrels()
            self.queries = {qid: self.queries[qid] for qid in self.qrels}
            logger.info(f"Loaded {len(self.queries)} {split.upper()} Queries.")
            logger.info(f"Query Example: {list(self.queries.values())[0]}")

        return self.corpus, self.queries, self.qrels

    def load_corpus(self) -> Dict[str, Dict[str, str]]:
        self.check(fIn=self.corpus_file, ext="jsonl")

        if not len(self.corpus):
            logger.info("Loading Corpus...")
            self._load_corpus()
            logger.info(f"Loaded {len(self.corpus)} Documents.")
            logger.info(f"Doc Example: {list(self.corpus.values())[0]}")

        return self.corpus

    def _load_corpus(self):
        num_lines = sum(1 for i in open(self.corpus_file, 'rb'))
        with open(self.corpus_file, encoding='utf8') as fIn:
            for line in tqdm(fIn, total=num_lines):
                line = json.loads(line)
                self.corpus[line.get("_id")] = {
                    "text": line.get("text"),
                    "title": line.get("title"),
                }

    def _load_queries(self):
        with open(self.query_file, encoding='utf8') as fIn:
            for line in fIn:
                line = json.loads(line)
                self.queries[line.get("_id")] = line.get("text")

    def _load_qrels(self):
        reader = csv.reader(open(self.qrels_file, encoding="utf-8"),
                            delimiter="\t", quoting=csv.QUOTE_MINIMAL)
        next(reader)

        for id, row in enumerate(reader):
            query_id, corpus_id, score = row[0], row[1], int(row[2])

            if query_id not in self.qrels:
                self.qrels[query_id] = {corpus_id: score}
            else:
                self.qrels[query_id][corpus_id] = score



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Use faiss to search clip embeddings
"""
import base64
import json
import os
from io import BytesIO
from pathlib import Path
from typing import List, Optional, Union

import faiss
import fire
import numpy as np
import pandas as pd
import requests
from PIL import Image
from loguru import logger
from pydantic import BaseModel, Field
from requests.exceptions import RequestException

from similarities.clip_module import ClipModule
from similarities.utils.util import cos_sim


def preprocess_image(image_input: Union[str, np.ndarray, bytes]) -> Image.Image:
    """
    Process image input to Image.Image object, supporting all formats that PIL can handle.
    """
    try:
        if isinstance(image_input, str):
            # Try to open image from http or local file path
            if image_input.startswith('http'):
                response = requests.get(image_input, stream=True)
                response.raise_for_status()  # Raises HTTPError if the HTTP request returned an unsuccessful status code
                return Image.open(response.raw)
            elif os.path.isfile(image_input):
                return Image.open(image_input)
            else:
                raise ValueError(f"Image path does not exist or is not a file: {image_input}")
        elif isinstance(image_input, np.ndarray):
            # Convert numpy array to image
            return Image.fromarray(image_input)
        elif isinstance(image_input, bytes):
            # Open image from byte data
            img_data = BytesIO(base64.b64decode(image_input))
            return Image.open(img_data)
        else:
            raise ValueError(f"Unsupported image input type, image input: {image_input}")
    except Exception as e:
        # Catch any exception that PIL might raise and re-raise it to the caller
        raise ValueError(f"An error occurred while processing the image input: {e}")
        

def clip_embedding(
        input_dir: str,
        chunk_size: int = 10000,
        image_embeddings_dir: Optional[str] = 'clip_engine/image_emb/',
        text_embeddings_dir: Optional[str] = 'clip_engine/text_emb/',
        corpus_dir: str = 'clip_engine/corpus/',
        model_name: str = "OFA-Sys/chinese-clip-vit-base-patch16",
        batch_size: int = 32,
        enable_image: bool = True,
        image_column_name: str = 'image_path',
        enable_text: bool = False,
        text_column_name: str = 'text',
        target_devices: List[str] = None,
        normalize_embeddings: bool = False,
        **kwargs,
):
    """
    Embedding text and image with clip model
    :param input_dir: input dir, support tsv/csv/txt files
    :param chunk_size: chunk size to read the input file
    :param image_embeddings_dir: save image embeddings dir
    :param text_embeddings_dir: save text embeddings dir
    :param corpus_dir: save corpus dir
    :param model_name: clip model name
    :param batch_size: batch size to compute embeddings, default 32
    :param enable_image: enable image embedding, default True
    :param image_column_name: image column name from input file, default image_path
    :param enable_text: enable text embedding, default False
    :param text_column_name: text column name from input file, default text
    :param target_devices: pytorch target devices, e.g. ['cuda:0', cuda:1]
        If None, all available CUDA devices will be used
    :param normalize_embeddings: whether to normalize embeddings before saving
    :param kwargs: read_csv kwargs
    :return: None, save embeddings to image_embeddings_dir and text_embeddings_dir
    """
    assert enable_image or enable_text, "enable_image and enable_text should not be both False"
    input_files = [f for f in os.listdir(input_dir) if f.endswith((".tsv", ".csv", ".txt"))]
    assert len(input_files) > 0, f"input_dir {input_dir} has no tsv/csv/txt files"
    logger.info(f"Start embedding, input files: {input_files}")
    model = ClipModule(model_name_or_path=model_name)
    logger.info(f'Load model success. model: {model_name}')

    for i, file in enumerate(input_files):
        logger.debug(f"Processing file {i + 1}/{len(input_files)}: {file}")
        output_file_counter = 0
        input_path = os.path.join(input_dir, file)
        # Read the input file in chunks
        for chunk_df in pd.read_csv(input_path, chunksize=chunk_size, **kwargs):
            if enable_image:
                images = chunk_df[image_column_name].tolist()

                images = [preprocess_image(img) for img in images]
                pool = model.start_multi_process_pool(target_devices=target_devices)
                # Compute the embeddings using the multi processes pool
                image_emb = model.encode_multi_process(
                    images,
                    pool,
                    batch_size=batch_size,
                    normalize_embeddings=normalize_embeddings
                )
                model.stop_multi_process_pool(pool)
                os.makedirs(image_embeddings_dir, exist_ok=True)
                image_embeddings_file = os.path.join(image_embeddings_dir, f"part-{output_file_counter:05d}.npy")
                np.save(image_embeddings_file, image_emb)
                logger.debug(f"Embeddings computed. Shape: {image_emb.shape}, saved to {image_embeddings_file}")
            if enable_text:
                texts = chunk_df[text_column_name].tolist()

                pool = model.start_multi_process_pool(target_devices=target_devices)
                # Compute the embeddings using the multi processes pool
                text_emb = model.encode_multi_process(
                    texts,
                    pool,
                    batch_size=batch_size,
                    normalize_embeddings=normalize_embeddings
                )
                model.stop_multi_process_pool(pool)
                os.makedirs(text_embeddings_dir, exist_ok=True)
                text_embeddings_file = os.path.join(text_embeddings_dir, f"part-{output_file_counter:05d}.npy")
                np.save(text_embeddings_file, text_emb)
                logger.debug(f"Embeddings computed. Shape: {text_emb.shape}, saved to {text_embeddings_file}")

            # Save corpus to Parquet file
            os.makedirs(corpus_dir, exist_ok=True)
            corpus_file = os.path.join(corpus_dir, f"part-{output_file_counter:05d}.parquet")
            chunk_df.to_parquet(corpus_file, index=False)
            logger.debug(f"Corpus data saved to {corpus_file}")
            output_file_counter += 1
    logger.info(f"Embedding done, saved image emb to {image_embeddings_dir} and text emb to {text_embeddings_dir}")


def clip_index(
        image_embeddings_dir: Optional[str] = None,
        text_embeddings_dir: Optional[str] = None,
        image_index_dir: Optional[str] = "clip_engine/image_index/",
        text_index_dir: Optional[str] = "clip_engine/text_index/",
        index_name: str = "faiss.index",
        max_index_memory_usage: str = "4G",
        current_memory_available: str = "16G",
        use_gpu: bool = False,
        nb_cores: Optional[int] = None,
):
    """
    Build indexes from embeddings using autofaiss
    :param image_embeddings_dir: image embeddings dir, required for image search
    :param text_embeddings_dir: optional, text embeddings dir
    :param image_index_dir: folder to save image index dir
    :param text_index_dir: folder to save text index dir
    :param index_name: index name to save, default 'faiss.index'
    :param max_index_memory_usage: Maximum size allowed for the index, this bound is strict
    :param current_memory_available: Memory available on the machine creating the index, having more memory is a boost
        because it reduces the swipe between RAM and disk.
    :param use_gpu: whether to use gpu, default False
    :param nb_cores: Number of cores to use. Will try to guess the right number if not provided
    :return: None, save index to image_index_dir and text_index_dir
    """
    try:
        from autofaiss import build_index  # pylint: disable=import-outside-toplevel
    except ImportError:
        raise ImportError("autofaiss is required for building indexes. Install it with: pip install similarities[index]")

    if image_embeddings_dir and os.path.exists(image_embeddings_dir):
        logger.debug(f"Starting build index from {image_embeddings_dir}")
        logger.debug(
            f"Embedding path exist, building index "
            f"using embeddings {image_embeddings_dir} ; saving in {image_index_dir}"
        )
        os.makedirs(image_index_dir, exist_ok=True)
        index_file = os.path.join(image_index_dir, index_name)
        index_infos_path = os.path.join(image_index_dir, index_name + ".json")
        try:
            build_index(
                embeddings=image_embeddings_dir,
                index_path=index_file,
                index_infos_path=index_infos_path,
                max_index_memory_usage=max_index_memory_usage,
                current_memory_available=current_memory_available,
                nb_cores=nb_cores,
                use_gpu=use_gpu,
            )
            logger.info(f"Index {image_embeddings_dir} done, saved in {index_file}, index infos in {index_infos_path}")
        except Exception as e:  # pylint: disable=broad-except
            logger.error(f"Index {image_embeddings_dir} failed, {e}")
            raise e
    else:
        logger.warning(f"Embeddings dir {image_embeddings_dir} not exist")

    if text_embeddings_dir and os.path.exists(text_embeddings_dir):
        logger.debug(f"Starting build index from {text_embeddings_dir}")
        logger.debug(
            f"Embedding path exist, building index "
            f"using embeddings {text_embeddings_dir} ; saving in {text_index_dir}"
        )
        os.makedirs(text_index_dir, exist_ok=True)
        index_file = os.path.join(text_index_dir, index_name)
        index_infos_path = os.path.join(text_index_dir, index_name + ".json")
        try:
            build_index(
                embeddings=text_embeddings_dir,
                index_path=index_file,
                index_infos_path=index_infos_path,
                max_index_memory_usage=max_index_memory_usage,
                current_memory_available=current_memory_available,
                nb_cores=nb_cores,
                use_gpu=use_gpu,
            )
            logger.info(f"Index {text_embeddings_dir} done, saved in {index_file}, index infos in {index_infos_path}")
        except Exception as e:  # pylint: disable=broad-except
            logger.error(f"Index {text_embeddings_dir} failed, {e}")
            raise e
    else:
        logger.warning(f"Embeddings dir {text_embeddings_dir} not exist")


def batch_search_index(
        queries,
        model,
        faiss_index,
        df,
        num_results,
        threshold,
        debug=False,
):
    """
    Search index with image inputs or image paths (batch search)
    :param queries: list of image paths or list of image inputs or texts or embeddings
    :param model: CLIP model
    :param faiss_index: faiss index
    :param df: corpus dataframe
    :param num_results: int, number of results to return
    :param threshold: float, threshold to return results
    :param debug: bool, whether to print debug info, default True
    :return: search results
    """
    result = []
    if isinstance(queries, np.ndarray):
        if queries.size == 0:
            return result
        query_features = queries
    else:
        if not queries:
            return result
        query_features = model.encode(queries, normalize_embeddings=True, convert_to_numpy=True)

    if query_features.shape[0] > 0:
        query_features = query_features.astype(np.float32)
        if threshold is not None:
            _, D, I = faiss_index.range_search(query_features, threshold)
        else:
            D, I = faiss_index.search(query_features, num_results)
        for query, d, i in zip(queries, D, I):
            # Sorted faiss search result with distance
            text_scores = []
            for ed, ei in zip(d, i):
                # Convert to json, avoid float values error
                item = df.iloc[ei].to_json(force_ascii=False)
                if debug:
                    logger.debug(f"query: {query}, Found: {item}, similarity: {ed}, id: {ei}")
                text_scores.append((item, float(ed), int(ei)))
            # Sort by score desc
            query_result = sorted(text_scores, key=lambda x: x[1], reverse=True)
            result.append(query_result)
    return result


def clip_filter(
        texts: Optional[List[str]] = None,
        images: Optional[List[str]] = None,
        embeddings: Optional[Union[np.ndarray, List[str]]] = None,
        output_file: str = "outputs/result.jsonl",
        model_name: str = "OFA-Sys/chinese-clip-vit-base-patch16",
        index_dir: str = 'clip_engine/image_index/',
        index_name: str = "faiss.index",
        corpus_dir: str = 'clip_engine/corpus/',
        num_results: int = 10,
        threshold: Optional[float] = None,
        device: Optional[str] = None,
        debug: bool = False,
):
    """
    Entry point of clip filter, batch search index
    :param texts: optional, list of texts
    :param images: optional, list of image paths or list of image inputs
    :param embeddings: optional, list of embeddings
    :param output_file: output file path, default outputs/result.json
    :param model_name: clip model name
    :param index_dir: index dir, saved by clip_index, default clip_engine/image_index/
    :param index_name: index name, default `faiss.index`
    :param corpus_dir: corpus dir, saved by clip_embedding, default clip_engine/corpus/
    :param num_results: int, number of results to return
    :param threshold: float, threshold to return results
    :param device: pytorch device, e.g. 'cuda:0'
    :param debug: bool, whether to print debug info, default False
    :return: batch search results
    """
    if texts is None and images is None and embeddings is None:
        raise ValueError("must fill one of texts, images and embeddings input")
    index_file = os.path.join(index_dir, index_name)
    assert os.path.exists(index_file), f"index file {index_file} not exist"
    faiss_index = faiss.read_index(index_file)
    model = ClipModule(model_name_or_path=model_name, device=device)
    df = pd.concat(pd.read_parquet(parquet_file) for parquet_file in sorted(Path(corpus_dir).glob("*.parquet")))
    logger.info(f'Load success. model: {model_name}, index: {faiss_index}, corpus size: {len(df)}')

    queries = None
    if texts is not None and len(texts) > 0:
        queries = texts
        logger.debug(f"Query: texts size {len(texts)}")
    elif images is not None and len(images) > 0:
        queries = [preprocess_image(img) for img in images]
        logger.debug(f"Query: images size {len(images)}")
    elif embeddings is not None:
        queries = embeddings
        if isinstance(queries, list):
            queries = np.array(queries, dtype=np.float32)
        if len(queries.shape) == 1:
            queries = np.expand_dims(queries, axis=0)
        logger.debug(f"Query: embeddings shape {queries.shape}")
    result = batch_search_index(queries, model, faiss_index, df, num_results, threshold, debug=debug)
    # Save results
    if output_file:
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, 'w', encoding='utf-8') as f:
            if texts:
                for q, sorted_text_scores in zip(texts, result):
                    json.dump(
                        {'text': q,
                         'results': [{'item': i, 'similarity': j, 'id': k} for i, j, k in sorted_text_scores]},
                        f,
                        ensure_ascii=False,
                    )
                    f.write('\n')
                logger.info(f"Query texts size: {len(texts)}, saved result to {output_file}")
            elif images:
                for q, sorted_text_scores in zip(images, result):
                    json.dump(
                        {'image': q,
                         'results': [{'item': i, 'similarity': j, 'id': k} for i, j, k in sorted_text_scores]},
                        f,
                        ensure_ascii=False,
                    )
                    f.write('\n')
                logger.info(f"Query images size: {len(images)}, saved result to {output_file}")
            elif embeddings is not None:
                for q, sorted_text_scores in zip(queries, result):
                    json.dump(
                        {'emb': q.tolist(),
                         'results': [{'item': i, 'similarity': j, 'id': k} for i, j, k in sorted_text_scores]},
                        f,
                        ensure_ascii=False,
                    )
                    f.write('\n')
                logger.info(f"Query embeddings size: {len(embeddings)}, saved result to {output_file}")
    return result


class Item(BaseModel):
    input: str = Field(..., max_length=512)


class ClipItem(BaseModel):
    text: Optional[str] = Field(None, max_length=512)
    image: Optional[str] = None


class SearchItem(BaseModel):
    text: Optional[str] = Field(None, max_length=512)
    image: Optional[str] = None
    emb: Optional[List[float]] = None


def clip_server(
        model_name: str = "OFA-Sys/chinese-clip-vit-base-patch16",
        index_dir: str = 'clip_engine/image_index/',
        index_name: str = "faiss.index",
        corpus_dir: str = 'clip_engine/corpus/',
        num_results: int = 10,
        threshold: Optional[float] = None,
        device: Optional[str] = None,
        host: str = "0.0.0.0",
        port: int = 8002,
        debug: bool = False,
):
    """
    Main entry point of clip search backend, start the server
    :param model_name: clip model name
    :param index_dir: index dir, saved by clip_index, default clip_engine/image_index/
    :param index_name: index name, default `faiss.index`
    :param corpus_dir: corpus dir, saved by clip_embedding, default clip_engine/corpus/
    :param num_results: int, number of results to return
    :param threshold: float, threshold to return results
    :param device: pytorch device, e.g. 'cuda:0'
    :param host: server host, default '0.0.0.0'
    :param port: server port, default 8002
    :param debug: bool, whether to print debug info, default False
    :return: None, start the server
    """
    import uvicorn
    from fastapi import FastAPI
    from starlette.middleware.cors import CORSMiddleware

    logger.info("starting boot of clip server")
    index_file = os.path.join(index_dir, index_name)
    assert os.path.exists(index_file), f"index file {index_file} not exist"
    faiss_index = faiss.read_index(index_file)
    model = ClipModule(model_name_or_path=model_name, device=device)
    df = pd.concat(pd.read_parquet(parquet_file) for parquet_file in sorted(Path(corpus_dir).glob("*.parquet")))
    logger.info(f'Load model success. model: {model_name}, index: {faiss_index}, corpus size: {len(df)}')

    # define the app
    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"])

    @app.get('/')
    async def index():
        return {"message": "index, docs url: /docs"}

    @app.post('/emb')
    async def emb(item: ClipItem):
        try:
            if item.text is not None:
                q = [item.text]
            elif item.image is not None:
                q = [preprocess_image(item.image)]
            else:
                raise ValueError("item should have text or image")
            embeddings = model.encode(q)
            result_dict = {'emb': embeddings.tolist()[0]}
            logger.debug(f"Successfully get embeddings, res shape: {embeddings.shape}")
            return result_dict
        except Exception as e:
            logger.error(e)
            return {'status': False, 'msg': e}, 400

    @app.post('/similarity')
    async def similarity(item1: ClipItem, item2: ClipItem):
        try:
            if item1.text is not None:
                q1 = item1.text
            elif item1.image is not None:
                q1 = preprocess_image(item1.image)
            else:
                raise ValueError("item1 should have text or image")
            if item2.text is not None:
                q2 = item2.text
            elif item2.image is not None:
                q2 = preprocess_image(item2.image)
            else:
                raise ValueError("item2 should have text or image")
            emb1 = model.encode(q1)
            emb2 = model.encode(q2)
            sim_score = cos_sim(emb1, emb2).tolist()[0][0]
            result_dict = {'similarity': sim_score}
            logger.debug(f"Successfully get similarity score, res: {sim_score}")
            return result_dict
        except Exception as e:
            logger.error(e)
            return {'status': False, 'msg': e}, 400

    @app.post('/search')
    async def search(item: SearchItem):
        try:
            if item.text is not None and len(item.text) > 0:
                q = [item.text]
                logger.debug(f"query: text {item.text}")
            elif item.image is not None and len(item.image) > 0:
                q = [preprocess_image(item.image)]
                logger.debug(f"query: image {item.image}")
            elif item.emb is not None:
                q = item.emb
                logger.debug(f"query: emb size {len(item.emb)}")
                if isinstance(q, list):
                    q = np.array(q, dtype=np.float32)
                if len(q.shape) == 1:
                    q = np.expand_dims(q, axis=0)
            else:
                raise ValueError("item should have text or image or emb")
            results = batch_search_index(q, model, faiss_index, df, num_results, threshold, debug=debug)
            # batch search result, input is one, need return the first
            sorted_text_scores = results[0]
            result_dict = {'result': sorted_text_scores}
            logger.debug(f"Successfully search done, res size: {len(sorted_text_scores)}")
            return result_dict
        except Exception as e:
            logger.error(f"search error: {e}")
            return {'status': False, 'msg': e}, 400

    logger.info("Server starting!")
    uvicorn.run(app, host=host, port=port)


class ClipClient:
    def __init__(self, base_url: str = "http://0.0.0.0:8002", timeout: int = 30):
        self.base_url = base_url
        self.timeout = timeout

    def _post(self, endpoint: str, data: dict) -> dict:
        try:
            response = requests.post(f"{self.base_url}/{endpoint}", json=data, timeout=self.timeout)
            response.raise_for_status()
            return response.json()
        except RequestException as e:
            logger.error(f"Request failed: {e}")
            return {}

    def get_emb(self, text: Optional[str] = None, image: Optional[str] = None) -> List[float]:
        try:
            data = {
                "text": text,
                "image": image,
            }
            response = self._post("emb", data)
            return response.get("emb", [])
        except Exception as e:
            logger.error(e)
            return []

    def get_similarity(self, item1: ClipItem, item2: ClipItem) -> float:
        try:
            data = {"item1": item1.dict(), "item2": item2.dict()}
            response = self._post("similarity", data)
            return response.get("similarity", 0.0)
        except Exception as e:
            logger.error(f"Error: {e}")
            return 0.0

    def search(
            self,
            text: Optional[str] = None,
            image: Optional[str] = None,
            emb: Optional[List[float]] = None
    ):
        try:
            data = {
                "text": text,
                "image": image,
                "emb": emb
            }
            response = self._post("search", data)
            return response.get("result", [])
        except Exception as e:
            logger.error(f"Error: {e}")
            return []


def main():
    """Main entry point"""
    fire.Fire(
        {
            "clip_embedding": clip_embedding,
            "clip_index": clip_index,
            "clip_filter": clip_filter,
            "clip_server": clip_server,
        }
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:

Compute similarity:
1. Compute the similarity between two sentences
2. Retrieves most similar sentence of a query against a corpus of documents.
"""

from typing import List, Union, Dict
from PIL import Image


class SimilarityABC:
    """
    Interface for similarity compute and search.

    In all instances, there is a corpus against which we want to perform the similarity search.
    For each similarity search, the input is a document or a corpus, and the output are the similarities
    to individual corpus documents.
    """

    def add_corpus(self, corpus: Union[List, Dict]):
        """
        Extend the corpus with new documents.
        corpus : list of str
        """
        raise NotImplementedError("cannot instantiate Abstract Base Class")

    def similarity(self, a: Union[str, Image.Image, List], b: Union[str, Image.Image, List]):
        """
        Compute similarity between two texts.
        :param a: list of str or str
        :param b: list of str or str
        :param score_function: function to compute similarity, default cos_sim
        :return: similarity score, torch.Tensor, Matrix with res[i][j] = cos_sim(a[i], b[j])
        """
        raise NotImplementedError("cannot instantiate Abstract Base Class")

    def distance(self, a: Union[str, Image.Image, List], b: Union[str, Image.Image, List]):
        """Compute cosine distance between two texts."""
        raise NotImplementedError("cannot instantiate Abstract Base Class")

    def most_similar(self, queries: Union[str, Image.Image, List, Dict], topn: int = 10) -> List[List[Dict]]:
        """
        Find the topn most similar texts to the query against the corpus.
        :param queries: Dict[int(query_id), str(query_text)] or List[str] or str
        :param topn: int
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing similarity scores.
        """
        raise NotImplementedError("cannot instantiate Abstract Base Class")

    def search(self, queries: Union[str, Image.Image, List, Dict], topn: int = 10) -> List[List[Dict]]:
        """
        Find the topn most similar texts to the query against the corpus.
        :param queries: Dict[int(query_id), str(query_text)] or List[str] or str
        :param topn: int
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing similarity scores.
        """
        return self.most_similar(queries, topn=topn)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
copy from https://github.com/UKPLab/sentence-transformers/blob/master/sentence_transformers/util.py
"""

import heapq
import queue
from typing import Union, List, Dict

import numpy as np
import torch
import torch.nn.functional


def cos_sim(a: Union[torch.Tensor, np.ndarray], b: Union[torch.Tensor, np.ndarray]):
    """
    Computes the cosine similarity cos_sim(a[i], b[j]) for all i and j.
    :return: Matrix with res[i][j]  = cos_sim(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    a_norm = torch.nn.functional.normalize(a, p=2, dim=1)
    b_norm = torch.nn.functional.normalize(b, p=2, dim=1)
    return torch.mm(a_norm, b_norm.transpose(0, 1))


def dot_score(a: Union[torch.Tensor, np.ndarray], b: Union[torch.Tensor, np.ndarray]):
    """
    Computes the dot-product dot_prod(a[i], b[j]) for all i and j.
    :return: Matrix with res[i][j]  = dot_prod(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    return torch.mm(a, b.transpose(0, 1))


def pairwise_dot_score(a: Union[torch.Tensor, np.ndarray], b: Union[torch.Tensor, np.ndarray]):
    """
    Computes the pairwise dot-product dot_prod(a[i], b[i])
    :return: Vector with res[i] = dot_prod(a[i], b[i])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    return (a * b).sum(dim=-1)


def pairwise_cos_sim(a: Union[torch.Tensor, np.ndarray], b: Union[torch.Tensor, np.ndarray]):
    """
   Computes the pairwise cossim cos_sim(a[i], b[i])
   :return: Vector with res[i] = cos_sim(a[i], b[i])
   """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    return pairwise_dot_score(normalize_embeddings(a), normalize_embeddings(b))


def normalize_embeddings(embeddings: torch.Tensor):
    """
    Normalizes the embeddings matrix, so that each sentence embedding has unit length
    """
    return torch.nn.functional.normalize(embeddings, p=2, dim=1)


def semantic_search(
        query_embeddings: Union[torch.Tensor, np.ndarray],
        corpus_embeddings: Union[torch.Tensor, np.ndarray],
        query_chunk_size: int = 100,
        corpus_chunk_size: int = 500000,
        top_k: int = 10,
        score_function=cos_sim
) -> List[List[Dict[str, Union[int, float]]]]:
    """
    This function performs a cosine similarity search between a list of query embeddings and corpus embeddings.
    It can be used for Information Retrieval / Semantic Search for corpora up to about 1 Million entries.

    :param query_embeddings: A 2-dimensional tensor with the query embeddings.
    :param corpus_embeddings: A 2-dimensional tensor with the corpus embeddings.
    :param query_chunk_size: Process 100 queries simultaneously. Increasing that value increases the speed, but
        requires more memory.
    :param corpus_chunk_size: Scans the corpus 100k entries at a time. Increasing that value increases the speed,
        but requires more memory.
    :param top_k: Retrieve top k matching entries.
    :param score_function: Funtion for computing scores. By default, cosine similarity.
    :return: List[List[Dict[str, Union[int, float]]]]: A list with one entry for each query. Each entry is a list of
        dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing cosine similarity scores.
    """

    if isinstance(query_embeddings, (np.ndarray, np.generic)):
        query_embeddings = torch.from_numpy(query_embeddings)
    elif isinstance(query_embeddings, list):
        query_embeddings = torch.stack(query_embeddings)

    if len(query_embeddings.shape) == 1:
        query_embeddings = query_embeddings.unsqueeze(0)

    if isinstance(corpus_embeddings, (np.ndarray, np.generic)):
        corpus_embeddings = torch.from_numpy(corpus_embeddings)
    elif isinstance(corpus_embeddings, list):
        corpus_embeddings = torch.stack(corpus_embeddings)

    # Check that corpus and queries are on the same device
    if corpus_embeddings.device != query_embeddings.device:
        query_embeddings = query_embeddings.to(corpus_embeddings.device)

    queries_result_list = [[] for _ in range(len(query_embeddings))]

    for query_start_idx in range(0, len(query_embeddings), query_chunk_size):
        # Iterate over chunks of the corpus
        for corpus_start_idx in range(0, len(corpus_embeddings), corpus_chunk_size):
            # Compute cosine similarity
            cos_scores = score_function(query_embeddings[query_start_idx:query_start_idx + query_chunk_size],
                                        corpus_embeddings[corpus_start_idx:corpus_start_idx + corpus_chunk_size])

            # Get top-k scores
            cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(
                cos_scores, min(top_k, len(cos_scores[0])), dim=1, largest=True, sorted=False)
            cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
            cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()

            for query_itr in range(len(cos_scores)):
                for sub_corpus_id, score in zip(cos_scores_top_k_idx[query_itr], cos_scores_top_k_values[query_itr]):
                    corpus_id = corpus_start_idx + sub_corpus_id
                    query_id = query_start_idx + query_itr
                    if len(queries_result_list[query_id]) < top_k:
                        heapq.heappush(queries_result_list[query_id], (
                            score, corpus_id))  # heaqp tracks the quantity of the first element in the tuple
                    else:
                        heapq.heappushpop(queries_result_list[query_id], (score, corpus_id))

    # change the data format and sort
    for query_id in range(len(queries_result_list)):
        for doc_itr in range(len(queries_result_list[query_id])):
            score, corpus_id = queries_result_list[query_id][doc_itr]
            queries_result_list[query_id][doc_itr] = {'corpus_id': corpus_id, 'score': score}
        queries_result_list[query_id] = sorted(queries_result_list[query_id], key=lambda x: x['score'], reverse=True)

    return queries_result_list


def paraphrase_mining_embeddings(
        embeddings: torch.Tensor,
        query_chunk_size: int = 5000,
        corpus_chunk_size: int = 100000,
        max_pairs: int = 500000,
        top_k: int = 100,
        score_function=cos_sim
):
    """
    Given a list of sentences / texts, this function performs paraphrase mining. It compares all sentences against all
    other sentences and returns a list with the pairs that have the highest cosine similarity score.

    :param embeddings: A tensor with the embeddings
    :param query_chunk_size: Search for most similar pairs for #query_chunk_size at the same time. Decrease, to lower memory footprint (increases run-time).
    :param corpus_chunk_size: Compare a sentence simultaneously against #corpus_chunk_size other sentences. Decrease, to lower memory footprint (increases run-time).
    :param max_pairs: Maximal number of text pairs returned.
    :param top_k: For each sentence, we retrieve up to top_k other sentences
    :param score_function: Function for computing scores. By default, cosine similarity.
    :return: Returns a list of triplets with the format [score, id1, id2]
    """

    top_k += 1  # A sentence has the highest similarity to itself. Increase +1 as in distinct pairs

    # Mine for duplicates
    pairs = queue.PriorityQueue()
    min_score = -1
    num_added = 0

    for corpus_start_idx in range(0, len(embeddings), corpus_chunk_size):
        for query_start_idx in range(0, len(embeddings), query_chunk_size):
            scores = score_function(embeddings[query_start_idx:query_start_idx + query_chunk_size],
                                    embeddings[corpus_start_idx:corpus_start_idx + corpus_chunk_size])

            scores_top_k_values, scores_top_k_idx = torch.topk(
                scores, min(top_k, len(scores[0])), dim=1, largest=True, sorted=False)
            scores_top_k_values = scores_top_k_values.cpu().tolist()
            scores_top_k_idx = scores_top_k_idx.cpu().tolist()

            for query_itr in range(len(scores)):
                for top_k_idx, corpus_itr in enumerate(scores_top_k_idx[query_itr]):
                    i = query_start_idx + query_itr
                    j = corpus_start_idx + corpus_itr

                    if i != j and scores_top_k_values[query_itr][top_k_idx] > min_score:
                        pairs.put((scores_top_k_values[query_itr][top_k_idx], i, j))
                        num_added += 1

                        if num_added >= max_pairs:
                            entry = pairs.get()
                            min_score = entry[0]

    # Get the pairs
    added_pairs = set()  # Used for duplicate detection
    pairs_list = []
    while not pairs.empty():
        score, i, j = pairs.get()
        sorted_i, sorted_j = sorted([i, j])

        if sorted_i != sorted_j and (sorted_i, sorted_j) not in added_pairs:
            added_pairs.add((sorted_i, sorted_j))
            pairs_list.append([score, i, j])

    # Highest scores first
    pairs_list = sorted(pairs_list, key=lambda x: x[0], reverse=True)
    return pairs_list


def community_detection(embeddings, threshold=0.75, min_community_size=10, batch_size=10000):
    """
    Function for Fast Community Detection
    Finds in the embeddings all communities, i.e. embeddings that are close (closer than threshold).
    Returns only communities that are larger than min_community_size. The communities are returned
    in decreasing order. The first element in each list is the central point in the community.
    """
    if not isinstance(embeddings, torch.Tensor):
        embeddings = torch.tensor(embeddings)

    threshold = torch.tensor(threshold, device=embeddings.device)

    extracted_communities = []

    # Maximum size for community
    min_community_size = min(min_community_size, len(embeddings))
    sort_max_size = min(max(2 * min_community_size, 50), len(embeddings))

    for start_idx in range(0, len(embeddings), batch_size):
        # Compute cosine similarity scores
        cos_scores = cos_sim(embeddings[start_idx:start_idx + batch_size], embeddings)

        # Minimum size for a community
        top_k_values, _ = cos_scores.topk(k=min_community_size, largest=True)

        # Filter for rows >= min_threshold
        for i in range(len(top_k_values)):
            if top_k_values[i][-1] >= threshold:
                new_cluster = []

                # Only check top k most similar entries
                top_val_large, top_idx_large = cos_scores[i].topk(k=sort_max_size, largest=True)

                # Check if we need to increase sort_max_size
                while top_val_large[-1] > threshold and sort_max_size < len(embeddings):
                    sort_max_size = min(2 * sort_max_size, len(embeddings))
                    top_val_large, top_idx_large = cos_scores[i].topk(k=sort_max_size, largest=True)

                for idx, val in zip(top_idx_large.tolist(), top_val_large):
                    if val < threshold:
                        break

                    new_cluster.append(idx)

                extracted_communities.append(new_cluster)

        del cos_scores

    # Largest cluster first
    extracted_communities = sorted(extracted_communities, key=lambda x: len(x), reverse=True)

    # Step 2) Remove overlapping communities
    unique_communities = []
    extracted_ids = set()

    for cluster_id, community in enumerate(extracted_communities):
        community = sorted(community)
        non_overlapped_community = []
        for idx in community:
            if idx not in extracted_ids:
                non_overlapped_community.append(idx)

        if len(non_overlapped_community) >= min_community_size:
            unique_communities.append(non_overlapped_community)
            extracted_ids.update(non_overlapped_community)

    unique_communities = sorted(unique_communities, key=lambda x: len(x), reverse=True)

    return unique_communities



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import sys
import unittest

sys.path.append('..')
from similarities import Similarity

m = Similarity()


class SimScoreTestCase(unittest.TestCase):

    def test_sim_diff(self):
        a = '研究团队面向国家重大战略需求追踪国际前沿发展借鉴国际人工智能研究领域的科研模式有效整合创新资源解决复'
        b = '英汉互译比较语言学'
        r = m.similarity(a, b)[0][0]
        r = float(r)
        print(a, b, r)
        self.assertTrue(abs(r - 0.4098) < 0.001)

    def test_empty(self):
        v = m.get_embeddings("This is test1")
        print(v[:10], v.shape)
        print(m.similarity("This is a test1", "that is a test5"))
        print(m.distance("This is a test1", "that is a test5"))
        print(m.most_similar("This is a test4"))
        r = m.most_similar('刘若英是演员', topn=10)
        print(r)
        self.assertEqual(len(r[0]), 0)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)
        self.assertEqual(len(r[0]), 0)

    def test_case(self):
        cases = [("牙疼有蛀牙怎么办", "手机前十名排行榜"),
                 ("小游戏,下载", "干洗机什么牌子的好"),
                 ("如何恢复微信里面的聊天记录", "吸氢气机"),
                 ("胃病症状的早期表现", "胃溃疡症状"), ]
        for i in cases:
            print(i)
            r = m.similarity(i[0], i[1])
            print(r)
        print("-" * 50)
        new_m = Similarity(model_name_or_path="sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2")
        for i in cases:
            print(i)
            r = new_m.similarity(i[0], i[1])
            print(r)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""

import base64
import sys
from io import BytesIO

import cv2
import numpy as np
import requests
from PIL import Image
from loguru import logger
from tqdm import tqdm


def is_link(s):
    return s is not None and s.startswith('http')


def img_decode(content: bytes):
    np_arr = np.frombuffer(content, dtype=np.uint8)
    return cv2.imdecode(np_arr, cv2.IMREAD_UNCHANGED)


def download_with_progressbar(url, save_path):
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        total_size_in_bytes = int(response.headers.get('content-length', 1))
        block_size = 1024  # 1 Kibibyte
        progress_bar = tqdm(
            total=total_size_in_bytes, unit='iB', unit_scale=True)
        with open(save_path, 'wb') as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
    else:
        logger.error("Something went wrong while downloading models")
        sys.exit(0)


def check_img(img):
    if isinstance(img, bytes):
        img = img_decode(img)
    if isinstance(img, str):
        # download net image
        if is_link(img):
            download_with_progressbar(img, 'tmp.jpg')
            img = 'tmp.jpg'
        image_file = img
        with open(image_file, 'rb') as f:
            img_str = f.read()
            img = img_decode(img_str)
        if img is None:
            try:
                buf = BytesIO()
                image = BytesIO(img_str)
                im = Image.open(image)
                rgb = im.convert('RGB')
                rgb.save(buf, 'jpeg')
                buf.seek(0)
                image_bytes = buf.read()
                data_base64 = str(base64.b64encode(image_bytes),
                                  encoding="utf-8")
                image_decode = base64.b64decode(data_base64)
                img_array = np.frombuffer(image_decode, np.uint8)
                img = cv2.imdecode(img_array, cv2.IMREAD_COLOR)
            except:
                logger.error("error in loading image:{}".format(image_file))
                return None
        if img is None:
            logger.error("error in loading image:{}".format(image_file))
            return None
    if isinstance(img, np.ndarray) and len(img.shape) == 2:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)

    return img


def alpha_to_color(img, alpha_color=(255, 255, 255)):
    if len(img.shape) == 3 and img.shape[2] == 4:
        B, G, R, A = cv2.split(img)
        alpha = A / 255

        R = (alpha_color[0] * (1 - alpha) + R * alpha).astype(np.uint8)
        G = (alpha_color[1] * (1 - alpha) + G * alpha).astype(np.uint8)
        B = (alpha_color[2] * (1 - alpha) + B * alpha).astype(np.uint8)

        img = cv2.merge((B, G, R))
    return img


def preprocess_image(img, alpha_color=(255, 255, 255)):
    """
    preprocess image
    :param img:
    :param alpha_color:
    :return:
    """
    img = check_img(img)
    if img is None:
        return None
    img = alpha_to_color(img, alpha_color)
    return img



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
import os
from typing import List, Union, Dict

from loguru import logger

from similarities.bert_similarity import BertSimilarity


class AnnoySimilarity(BertSimilarity):
    """
    Computes cosine similarities between word embeddings and retrieves most
    similar query for a given docs with Annoy.
    """

    def __init__(
            self,
            corpus: Union[List[str], Dict[str, str]] = None,
            model_name_or_path="shibing624/text2vec-base-chinese",
            n_trees: int = 256,
            device: str = None
    ):
        super().__init__(corpus, model_name_or_path, device=device)
        self.index = None
        self.embedding_size = self.get_sentence_embedding_dimension()
        self.n_trees = n_trees
        if corpus is not None and self.corpus_embeddings:
            self.build_index()

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: {self.sentence_model}"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def create_index(self):
        """Create Annoy Index."""
        try:
            from annoy import AnnoyIndex
        except ImportError:
            raise ImportError("Annoy is not installed. Please install it first, e.g. with `pip install annoy`.")

        # Creating the annoy index
        self.index = AnnoyIndex(self.embedding_size, 'angular')
        logger.debug(f"Init Annoy index, embedding_size: {self.embedding_size}")

    def build_index(self):
        """Build Annoy index after add new documents."""
        self.create_index()
        logger.debug(f"Building index with {self.n_trees} trees.")

        for i in range(len(self.corpus_embeddings)):
            self.index.add_item(i, self.corpus_embeddings[i])
        self.index.build(self.n_trees)

    def save_index(self, index_path: str = "annoy_index.bin"):
        """Save the annoy index to disk."""
        if index_path:
            if self.index is None:
                self.build_index()
            self.index.save(index_path)
            corpus_emb_json_path = index_path + ".jsonl"
            super().save_corpus_embeddings(corpus_emb_json_path)
            logger.info(f"Saving Annoy index to: {index_path}, corpus embedding to: {corpus_emb_json_path}")
        else:
            logger.warning("No index path given. Index not saved.")

    def load_index(self, index_path: str = "annoy_index.bin"):
        """Load Annoy Index from disc."""
        if index_path and os.path.exists(index_path):
            corpus_emb_json_path = index_path + ".jsonl"
            logger.info(f"Loading index from: {index_path}, corpus embedding from: {corpus_emb_json_path}")
            super().load_corpus_embeddings(corpus_emb_json_path)
            if self.index is None:
                self.create_index()
            self.index.load(index_path)
        else:
            logger.warning("No index path given. Index not loaded.")

    def most_similar(
            self,
            queries: Union[str, List[str], Dict[str, str]],
            topn: int = 10,
            score_function: str = "cos_sim",
            **kwargs
    ) -> List[List[Dict]]:
        """
        Find the topn most similar texts to the query against the corpus, using Annoy.
        :param queries: list of str or str
        :param topn: int
        :score_function: str, cosine similarity function
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing cosine similarity scores.
        """
        result = []
        if self.corpus_embeddings and self.index is None:
            logger.warning(f"No index found. Please add corpus and build index first, e.g. with `build_index()`."
                           f"Now returning slow search result.")
            return super().most_similar(queries, topn, score_function=score_function)
        if not self.corpus_embeddings:
            logger.error("No corpus_embeddings found. Please add corpus first, e.g. with `add_corpus()`.")
            return result
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []
        queries_texts = list(queries.values())
        queries_embeddings = self.get_embeddings(queries_texts, **kwargs)
        # Annoy get_nns_by_vector can only search for one vector at a time
        for idx, (qid, query) in enumerate(queries.items()):
            q_res = []
            corpus_ids, distances = self.index.get_nns_by_vector(queries_embeddings[idx], topn, include_distances=True)
            for corpus_id, distance in zip(corpus_ids, distances):
                score = 1 - (distance ** 2) / 2
                q_res.append({
                    "corpus_id": corpus_id,
                    "corpus_doc": self.corpus.get(corpus_id),
                    "score": score
                })
            result.append(q_res)
        return result


class HnswlibSimilarity(BertSimilarity):
    """
    Computes cosine similarities between word embeddings and retrieves most
    similar query for a given docs with Hnswlib.
    """

    def __init__(
            self,
            corpus: Union[List[str], Dict[str, str]] = None,
            model_name_or_path="shibing624/text2vec-base-chinese",
            ef_construction: int = 400, M: int = 64, ef: int = 50,
            device: str = None,
    ):
        super().__init__(corpus, model_name_or_path, device=device)
        self.embedding_size = self.get_sentence_embedding_dimension()
        self.ef_construction = ef_construction
        self.M = M
        self.ef = ef
        self.index = None
        if corpus is not None and self.corpus_embeddings:
            self.build_index()

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: {self.sentence_model}"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def create_index(self):
        """Create Hnswlib Index."""
        try:
            import hnswlib
        except ImportError:
            raise ImportError("Hnswlib is not installed. Please install it first, e.g. with `pip install hnswlib`.")

        # Creating the hnswlib index
        self.index = hnswlib.Index(space='cosine', dim=self.embedding_size)
        self.index.init_index(max_elements=len(self.corpus_embeddings), ef_construction=self.ef_construction, M=self.M)
        # Controlling the recall by setting ef:
        self.index.set_ef(self.ef)  # ef should always be > top_k_hits
        logger.debug(f"Init Hnswlib index, embedding_size: {self.embedding_size}")

    def build_index(self):
        """Build Hnswlib index after add new documents."""
        # Init the HNSWLIB index
        self.create_index()
        logger.info(f"Building HNSWLIB index, max_elements: {len(self.corpus)}")
        logger.debug(f"Parameters Required: M: {self.M}")
        logger.debug(f"Parameters Required: ef_construction: {self.ef_construction}")
        logger.debug(f"Parameters Required: ef(>topn): {self.ef}")

        # Then we train the index to find a suitable clustering
        self.index.add_items(self.corpus_embeddings, list(range(len(self.corpus_embeddings))))

    def save_index(self, index_path: str = "hnswlib_index.bin"):
        """Save the index to disk."""
        if index_path:
            if self.index is None:
                self.build_index()
            self.index.save_index(index_path)
            corpus_emb_json_path = index_path + ".jsonl"
            super().save_corpus_embeddings(corpus_emb_json_path)
            logger.info(f"Saving hnswlib index to: {index_path}, corpus embedding to: {corpus_emb_json_path}")
        else:
            logger.error("No index path given. Index not saved.")

    def load_index(self, index_path: str = "hnswlib_index.bin"):
        """Load Index from disc."""
        if index_path and os.path.exists(index_path):
            corpus_emb_json_path = index_path + ".jsonl"
            logger.info(f"Loading index from: {index_path}, corpus embedding from: {corpus_emb_json_path}")
            super().load_corpus_embeddings(corpus_emb_json_path)
            if self.index is None:
                self.create_index()
            self.index.load_index(index_path)
        else:
            logger.warning("No index path given. Index not loaded.")

    def most_similar(
            self,
            queries: Union[str, List[str], Dict[str, str]],
            topn: int = 10,
            score_function: str = "cos_sim",
            **kwargs
    ) -> List[List[Dict]]:
        """
        Find the topn most similar texts to the query against the corpus, using Hnswlib.
        :param queries: list of str or str
        :param topn: int
        :score_function: str, cosine similarity function
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing cosine similarity scores.
        """
        result = []
        if self.corpus_embeddings and self.index is None:
            logger.warning(f"No index found. Please add corpus and build index first, e.g. with `build_index()`."
                           f"Now returning slow search result.")
            return super().most_similar(queries, topn, score_function=score_function)
        if not self.corpus_embeddings:
            logger.error("No corpus_embeddings found. Please add corpus first, e.g. with `add_corpus()`.")
            return result
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        queries_texts = list(queries.values())
        queries_embeddings = self.get_embeddings(queries_texts, **kwargs)
        # We use hnswlib knn_query method to find the top_k_hits
        corpus_ids, distances = self.index.knn_query(queries_embeddings, k=topn)
        # We extract corpus ids and scores for each query
        for i, (qid, query) in enumerate(queries.items()):
            hits = [{'corpus_id': id, 'corpus_doc': self.corpus.get(id),
                     'score': 1 - distance} for id, distance in zip(corpus_ids[i], distances[i])]
            hits = sorted(hits, key=lambda x: x['score'], reverse=True)
            result.append(hits)
        return result



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""

import sys
import unittest

sys.path.append('..')

from similarities.literal_similarity import (
    SimHashSimilarity,
    TfidfSimilarity,
    BM25Similarity,
    WordEmbeddingSimilarity,
    CilinSimilarity,
    HownetSimilarity,
    SameCharsSimilarity,
    SequenceMatcherSimilarity,
)

from similarities.utils.distance import hamming_distance


class LiteralCase(unittest.TestCase):
    def test_hamming_distance(self):
        text1 = '刘若英是个演员'
        text2 = '他唱歌很好听'
        m = SimHashSimilarity()
        seq1 = m.simhash(text1)
        seq2 = m.simhash(text2)
        print(seq1)
        print(seq2)
        r = 1.0 - hamming_distance(seq1, seq2) / 64
        print(hamming_distance(seq1, seq2))
        print(r)
        print(f"{text1} vs {text2} sim score {m.similarity(text1, text2)}")

        text1 = '刘若英是个演员'
        text2 = '他'
        m = SimHashSimilarity()
        seq1 = m.simhash(text1)
        seq2 = m.simhash(text2)
        print(seq1)
        print(seq2)
        print(f"{text1} vs {text2} sim score {m.similarity(text1, text2)}")

        text1 = '刘若英唱歌'
        text2 = '唱歌'
        m = SimHashSimilarity()
        seq1 = m.simhash(text1)
        seq2 = m.simhash(text2)
        print(seq1)
        print(seq2)
        print(f"{text1} vs {text2} sim score {m.similarity(text1, text2)}")

        text1 = '刘若'
        text2 = '刘若他他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听,他唱歌很好听?他唱歌很好听？他唱歌很好听。。'
        m = SimHashSimilarity()
        seq1 = m.simhash(text1)
        seq2 = m.simhash(text2)
        print(seq1)
        print(seq2)
        print(f"{text1} vs {text2} sim score {m.similarity(text1, text2)}")

        text1 = '刘若 他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听,他唱歌很好听?他唱歌很好听？他唱歌很好'
        text2 = '他他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听他唱歌很好听,他唱歌很好听?他唱歌很好听？他唱歌很好听。。'
        m = SimHashSimilarity()
        seq1 = m.simhash(text1)
        seq2 = m.simhash(text2)
        print(seq1)
        print(seq2)
        s = m.similarity(text1, text2)
        print(f"{text1} vs {text2} sim score {s}")
        self.assertTrue(s[0] > 0.5)

    def test_simhash(self):
        """test_simhash"""
        text1 = '刘若英是个演员'
        text2 = '他唱歌很好听'
        m = SimHashSimilarity()
        print(f"{text1} vs {text2} sim score {m.similarity(text1, text2)}")
        print(m.distance(text1, text2))
        r = m.most_similar('刘若英是演员')
        self.assertEqual(len(r[0]), 0)
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌']
        m.add_corpus(zh_list)
        r = m.most_similar('刘若英是演员', topn=2)
        print(r)
        self.assertEqual(len(r[0]), 2)

    def test_tfidf(self):
        """test_tfidf"""
        text1 = '刘若英是个演员'
        text2 = '他唱歌很好听'
        m = TfidfSimilarity()
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌', '我不是演员吗']
        m.add_corpus(zh_list)
        r = m.most_similar('刘若英是演员')
        print(r)
        self.assertEqual(len(r[0]), 4)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)

    def test_bm25(self):
        """test_bm25"""
        text1 = '刘若英是个演员'
        text2 = '他唱歌很好听'
        m = BM25Similarity()
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌', '我不是演员吗']
        m.add_corpus(zh_list)
        r = m.most_similar('刘若英是演员', topn=10)
        print(r)
        self.assertEqual(len(r[0]), 4)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)

    def test_word2vec(self):
        """test_word2vec"""
        text1 = '刘若英是个演员'
        text2 = '他唱歌很好听'
        list_of_corpus = ["This is a test1", "This is a test2", "This is a test3"]
        list_of_corpus2 = ["that is test4", "that is a test5", "that is a test6"]
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌', '刘若英是个演员', '演戏很好看的人']
        m = WordEmbeddingSimilarity(list_of_corpus)
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))
        m.add_corpus(list_of_corpus2 + zh_list)
        v = m._get_vector("This is a test1")
        print(v[:10], v.shape)
        r = m.most_similar('刘若英是演员', topn=4)
        print(r)
        self.assertEqual(len(r[0]), 4)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)

    def test_cilin(self):
        """test_cilin"""
        text1 = '周杰伦是一个歌手'
        text2 = '刘若英是个演员'
        m = CilinSimilarity()
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌']
        m.add_corpus(zh_list)
        r = m.most_similar('刘若英是演员', topn=3)
        print(r)
        self.assertEqual(len(r[0]), 3)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)

    def test_hownet(self):
        """test_cilin"""
        text1 = '周杰伦是一个歌手'
        text2 = '刘若英是个演员'
        m = HownetSimilarity()
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌']
        m.add_corpus(zh_list)
        r = m.most_similar('刘若英是演员', topn=2)
        print(r)
        self.assertEqual(len(r[0]), 2)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)

    def test_samechars(self):
        """test_samechars"""
        text1 = '周杰伦是一个歌手'
        text2 = '刘若英是个演员'
        m = SameCharsSimilarity()
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))

        text1 = '刘若英是演员'
        text2 = '刘若英是个演员'
        m = SameCharsSimilarity()
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌']
        m.add_corpus(zh_list)
        r = m.most_similar('刘若英是演员', topn=3)
        print(r)
        self.assertEqual(len(r[0]), 3)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)

    def test_seqmatcher(self):
        """test_seqmatcher"""
        text1 = """
        是缘分还是注定，选择了你——威然。家里以前就有台大众途观，算起来开了也有七八年了，一直很稳当大问题也没出过。多数是老爸和我在开，随着去年年底我家臭小子降临，我用车时间比老爸用的时间多多了，顿时感觉家里一台车有点周转不过来啦。于是就有着增加一台车的想法，经过全家人讨论来讨论去，最终决定买台MPV，其实我内心一直是想要台SUV的，奈何以一敌三完败，老爸老妈的想法很奇葩，说是要为二胎考虑，没有那么钱总让我去换新车开，老婆最直接说MPV坐起来舒服些，她要享受老板级的待遇，这样一来也注定了我的司机命。      话说选车过程也是很艰辛啊，左挑右选，各种对比试驾，最后还是选择了大众，因为有第一台的感受，觉得大众还是很靠谱的，再一个就是从安全的角度出发，大众也绝对的让我放心。老婆又不太懂，只说威然坐起来真的非常舒服，就这样把它定了下来。。。      废话不多说了，趁着等我家老婆大人的空隙，给它搞了个写真，做好准备，我要上图咯前俩依然采用得是大众家族化得设计，大灯与前中网得完美融合使得前脸看起来很舒服，一体式得中网加上镀铬饰条得装饰让前脸看起来紧致又有张力，符合高端商务得气质Viloran作为一款全新车型，从外观来看突破了大众家族设计元素。从前脸来看新车新颖、大气，车头格栅与前大灯融为一体，而且进气格栅面积更大更宽，整体感进一步提高。前保险杠贯穿式通风口，内部由镀铬填充，侧面看会显得很长，整体得造型也是增加了很多得线条，看起来更加得丰满车尾部的造型比较中庸很符合大众的形象。尾灯做的也是比较大的，很有辨识度。大众VILORAN 的车标位置也显示出这是一辆豪华车。大众下面带字母都是豪车。内饰的设计也是特别的豪华，精致，液晶仪表盘，多媒体大屏幕应有尽有！中控也是全触摸的，基本没有按键。第二排的双杯架和手机之家，下面还有2个SUB充电口。很人性化的设计。坐在车里看着窗外的风景疾驰而过，心里真的非常惬意！！
        
        """
        text2 = """
        samples_1/J.txt 让生命不留遗憾，我选择——威然，提车作业！。就在上个星期左右，我们终于把威然给提了回来。在这个城市和老乡打拼多年，总算有了我们自己的一席之地，大学毕业后，我和老乡选择留在了这个快节奏而又充满着机遇和挑战的城市，在这个城市我们从初踏入职场，到联合创业，有过太多的心酸、也有太多的欢乐，我们始终坚信通过我们的努力定会在这个城市绽放出属于我们的精彩，在这个城市我们都邂逅了我们的挚爱，公司业务不忙的空档，我们两家也经常组团旅游，但很少自驾出去过，出于公司业务和家庭出游的需要，我和老乡商量着准备入手一台商务车，预算定在了三四十万，我们也知道这个价格开的GL8居多，但我们都感觉GL8太商务了，我们想买个宜商宜家的车，业务需要，有时会机场、高铁站接客户，所以对第二排要求一定要舒适、最好带按摩功能，这样我们的客户旅途劳顿可以在车上短暂的休息，无疑会提高我们公司服务形象，而且有时候带父母出游的话，第二排乘坐舒适对他们也比较好，自己平常也很喜欢车，也经常看一些车评，当时就看到了大众新上了一款大型的MPV威然，于是对他很好奇。威然给我的第一感觉就是酷、帅气,试驾的感觉也很棒。去了两次4S店，第二次是带着家人去的，买车不是件小事必须得征得媳妇的同意才行，我老婆见了这辆车之后也是很激动，能够看得出她那种喜悦之情。不仅是车子的外观吸引我，宽敞的空间对于我们家来说适合六口人乘坐，内饰的设计很大气高端有档次,说心里话我对她的感觉胜过了GL8，豪华感爆棚。所以最终还是把他提回来了。。图片已删除前俩依然采用得是大众家族化得设计，大灯与前中网得完美融合使得前脸看起来很舒服，一体式得中网加上镀铬饰条得装饰让前脸看起来紧致又有张力，符合高端商务得气质图片已删除Viloran作为一款全新车型，从外观来看突破了大众家族设计元素。从前脸来看新车新颖、大气，车头格栅与前大灯融为一体，而且进气格栅面积更大更宽，整体感进一步提高。前保险杠贯穿式通风口，内部由镀铬填充，图片已删除LED前大灯，炯炯有神图片已删除轮毂很漂亮图片已删除大红色尾灯，由银色饰条将两侧的尾灯相连，想不引人注意都不行图片已删除侧面看会显得很长，整体得造型也是增加了很多得线条，看起来更加得丰满图片已删除车尾部的造型比较中庸很符合大众的形象。尾灯做的也是比较大的，很有辨识度。大众VILORAN的车标位置也显示出这是一辆豪华车。大众下面带字母都是豪车。图片已删除威然这款车的后备箱超级大，同级别里最大的，不好的一点就是后排座椅和后备箱不平。图片已删除打开进入到主驾驶舱，一股豪华气息迎面而来。凡是平时接触比较多的位置都是采用了皮质的包裹，很舒服。而且这个浅色的内饰我个人也是非常喜欢。图片已删除多功能方向盘图片已删除第一眼就能看见这个全液晶的仪表盘，很有档次。图片已删除内饰的设计也是特别的豪华，精致，液晶仪表盘，多媒体大屏幕应有尽有！图片已删除中控也是全触摸的，基本没有按键。图片已删除杯架和杯架前面的都是储物空间。图片已删除扶手箱空间也蛮大的图片已删除图片已删除第二排的座椅，看上去就很舒服。也是电动调节的！图片已删除图片已删除第二排的双杯架和手机之家，下面还有2个SUB充电口。很人性化的设计。图片已删除门板非常厚实图片已删除图片已删除图片已删除第二排中央的空调出风口，还有座椅通风加热，都是很贴心的功能。图片已删除图片已删除全景大天窗，还有谁……图片已删除坐在车里看着窗外的风景疾驰而过，心里真的非常惬意！！
        """
        m = SequenceMatcherSimilarity()
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))

        text1 = '刘若英是演员'
        text2 = '刘若英是个演员'
        m = SequenceMatcherSimilarity()
        print(m.similarity(text1, text2))
        print(m.distance(text1, text2))
        zh_list = ['刘若英是个演员', '他唱歌很好听', 'women喜欢这首歌']
        m.add_corpus(zh_list)
        r = m.most_similar('刘若英是演员', topn=3)
        print(r)
        self.assertEqual(len(r[0]), 3)
        r = m.most_similar(['刘若英是演员', '唱歌很好听'])
        print(r)
        self.assertEqual(len(r), 2)

    def test_bm25_hardcase(self):
        """test_bm25 hardcase"""
        m = BM25Similarity()
        zh_list = [
            "的可能。术后还需要坚持服用药物治疗。药物治疗建议采用以补益心肾、涤痰熄风,开窍定痫,活血化淤、平肝泻火的中药汤药进行治疗,调理身体状况,调理脏腑机能,固本培元,达到治愈的目的。中药副作用小,标本兼治，治愈后不易复发",
            "我是星期一做的子宫息肉手术是否有关系吗，女51岁因为最近有些不舒服，就去进行了检查，但是一查是这个病，请问：我是星期一做的子宫息肉手术是否有关系",
            "腥膻的食物，一定是不要喝酒的，还有就是一定要多喝水，可以多吃些蔬果。前列腺囊肿这个疾病的患者一定是要注意休息的，一定是要保证自己充足的睡眠，还有就是一定是不要熬夜的，也尽量是不要高强度地工作的，还有就是一定是不要给自己过多的心理压力的，要学会放松自己。掌握前列腺囊肿的症状表现能够及时的发现疾病，并且不会因为由于对疾病的不了解，从而产生忽视，以及错误治疗的情况，事实上，对于前列腺疾病来说，非常容易根据其症状表现",
            "汤含服，或者用青果，凤凰衣等常用药物进行煎服，有一定效果。，考虑是肌肉软组织疼痛，建议口服风疼片，或者舒筋丸治疗都可以的，另外如果看中医方便也可地中医，进行中药治疗效果很好，积极对症治疗，慢慢会改善的，祝你健康",
            '内科护理学的辅助检查有些什么？", "answer": "客观结构化临床考试"}\n{"question": "白消安注射液的用法用量',
            "为疾病的统称来说，则具有比较基本的症状表现，毕竟类型不同，但是有些方面则是相同的，另外，疾病的类型不同，主要就是根据病因所决定，这点需要明确，下面就前列腺囊肿的基本症状表现进行总结：第一、排尿异常，主要的表现是排尿的异常(尿频、尿急、尿等待、尿分叉、排尿无力)。有的会有射精无力的症状，但是早期不会有症状。第二、分泌异常。主要是前列腺液异常的分泌导致的，(排尿滴白、排便滴白、尿频",
            '但是一查是这个病，请问：我是星期一做的子宫息肉手术是否有关系吗", "answer": "宫颈息肉（很小）手术后需要有卧床休息,当然如果出血量并不多,可适当的去户外活动的,防止重体力劳动就可,防止增强腹压而致使大出血的再次发生.宫颈息肉是慢性宫颈炎的一种表现出,息肉的根部粘附于子宫颈口或宫颈管内,是宫颈粘膜在炎症的刺激下局部纤维化,并外突于宫颈外口而构成.由于息肉割除比较简单,对身体伤损性也不大',
            '对患者有一定的影响。所以患者应该注意自身护理。流产患者注意自身护理，例如当的进行饮食调理，注意保暖避免着凉。"}\n{"question": "膺窗穴的定位是什么?", "answer": "第3肋间隙，距前正中线4寸。"}\n{"question": "土源性线虫感染的多发地区是哪里？", "answer": "苏北地区；贵州省剑河县；西南贫困地区；桂东；江西省鄱阳湖区；江西省',
            'question": "膺窗穴的定位是什么?", "answer": "第3肋间隙，距前正中线4寸。"}\n{"question": "土源性线虫感染的多发地区是哪里',
            '背痛的放射治疗有些什么？", "answer": "单纯体外放射治疗"}\n{"question": "盆腔淤血症的症状',
            '应该是患阴道炎了，而且是细菌性阴道炎的可能性大，你可以用妇科洗剂擦洗外阴阴道，同时口服些妇科千金片或者花红颗粒，具体用药谨遵医嘱，如果效果不佳。较好到医院去做白带常规仔细检查，查明病因后必要时再出针对性的用药治疗。平常以清淡饮食为基础，尽量多喝水。白带增多患者可以在理解专业治疗的同时，在生活中用食疗做为辅助治疗，在日常生活饮食中需要有恰当饮食，防止刺激性食物。',
            '天雄的药用植物栽培"',
            '[3]\t "致病，如风寒、风热、风湿、风燥等。"}\n{"question": "ESD术的手术治疗有些什么？", "answer": "ESD术"}\n{"question": "为什么会得大脖子病", "answer": "甲状腺肿的病因还没完全清楚，情绪、药物、化学物质、放射线、遗传缺损、炎症、自身免疫等因素干扰甲状腺激素的合成、储存与释放，及血中存在刺激甲状腺生长的因子都可引起甲状腺肿。结节性甲状腺肿的CT特点为病灶多发、形态',
            '静脉曲张及腹壁静脉曲张等等。那么，静脉曲张的最好治疗方法都有哪些？下面为大家详细的介绍下相关的知识。1、中医中药疗法：这也是临床效果较为显著的静脉曲张治疗方法，使用疏通经络的药物改变血管内壁的弹性，快速修复受损的瓣膜，使凸起的团状、条索状血管团逐渐缩小、平滑，能从根本上调节身体的机能。2、下肢静脉曲张有穿弹力袜、注射硬化剂、手术剥除等治疗方法，深静脉瓣膜功能不全，可作瓣膜修复手术和腔镜下交通支结扎术等。下肢',
            '对患者有一定的影响。所以患者应该注意自身护理。流产患者注意自身护理，例如当的进行饮食调理，注意保暖避免着凉。',
            '膺窗穴的定位是什么?", "answer": "第3肋间隙，距前正中线4寸。"}\n{"question": "土源性线虫感染的多发地区是哪里？", "answer": "苏北地区；贵州省剑河县；西南贫困地区；桂东；江西省鄱阳湖区；江西省',
            'question": "梭形细胞RMS的多发群体是什么？", "answer": "儿童和成年人',
            '我想了解艾灸可以治疗宫颈糜烂吗？，我有轻度的宫颈糜烂，听人说可以用艾灸来治疗，艾灸可以治疗宫颈糜烂吗？", "answer": "宫颈糜烂这种疾病，决定以上这种物理治疗措施，可以起些配置治疗作用，但是不能替代药物',
            '盆腔淤血症是盆腔炎，有小腹痛，白带异常，尿频等症状。如有此病，你应该尽快去医院治疗，以免延误病情。平时做好妇科查体。治疗可以使用环丙沙星加甲硝唑。注意日常的卫生习惯，勤洗澡。盆腔患者应在平时生活中忌食白酒，白扁豆，鸭蛋，鸭肉；宜吃富含维生素的食物、富含有优质蛋白质的食物和膳食纤维类食物，如：鸡蛋，鸡肉，鸡心，鹌鹑蛋。忌吃腥发的食物，如：羊肉、鲢鱼、鲤鱼；忌吃不容易消化的食物，如：螺丝肉、豆芽；3、忌吃富含油脂的食物，如：肥肉、猪油、羊油。{"queson": "梭形细胞RMS的多发群体是什么？", "answer": "儿童和成年人"}{"question": "我想了解艾灸可以治疗宫颈糜烂吗？，我有轻度的宫颈糜烂，听人说可以用艾灸来治疗，艾灸可以治疗宫颈糜烂吗',
            '对患者有一定的影响。所以患者应该注意自身护理。流产患者注意自身护理，例如当的进行饮食调理，注意保暖避免着凉。"}\n{"question": "膺窗穴的定位是什么?", "answer": "第3肋间隙，距前正中线4寸。"}\n{"question": "土源性线虫感染的多发地区是哪里？", "answer": "苏北地区；贵州省剑河县；西南贫困地区；桂东；江西省鄱阳湖区；江西省',
        ]
        m.add_corpus(zh_list)
        q = ['梭形细胞RMS的多发群体是什么？', '膺窗穴的定位是什么?']
        for i in q:
            res = m.most_similar(i, topn=10)
            print('sim search: ', res)
            for c in res:
                print("search res:")
                print(f'\t{c}')
            print('-' * 50 + '\n')



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Vit Novotny <witiko@mail.muni.cz>, lhy<lhy_in_blcu@126.com>
@description:
Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html

This module provides classes that deal with sentence similarities from mean term vector.
Adjust the gensim similarities Index to compute sentence similarities.
"""

import json
import os
from typing import List, Union, Dict

import jieba
import jieba.analyse
import jieba.posseg
import numpy as np
from loguru import logger
from tqdm import tqdm

from similarities.similarity import SimilarityABC
from similarities.utils.distance import string_hash, hamming_distance, longest_match_size
from similarities.utils.rank_bm25 import BM25Okapi
from similarities.utils.tfidf import TFIDF, load_stopwords, default_stopwords_file
from similarities.utils.util import cos_sim, semantic_search

pwd_path = os.path.abspath(os.path.dirname(__file__))


class SimHashSimilarity(SimilarityABC):
    """
    Compute SimHash similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    """

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None):
        self.corpus = {}

        self.corpus_embeddings = []
        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: SimHash"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : list of str or dict of str
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)
        corpus_texts = list(corpus_new.values())
        corpus_embeddings = []
        for sentence in tqdm(corpus_texts, desc="Computing corpus SimHash"):
            corpus_embeddings.append(self.simhash(sentence))
        if self.corpus_embeddings:
            self.corpus_embeddings += corpus_embeddings
        else:
            self.corpus_embeddings = corpus_embeddings
        logger.debug(f"Add {len(corpus_new)} docs, total: {len(self.corpus)}, emb size: {len(self.corpus_embeddings)}")

    def simhash(self, sentence: str):
        """
        Compute SimHash for a given text.
        :param sentence: str
        :return: hash code
        """
        seg = jieba.cut(sentence)
        key_word = jieba.analyse.extract_tags('|'.join(seg), topK=None, withWeight=True, allowPOS=())
        # 先按照权重排序，再按照词排序
        key_list = []
        for feature, weight in key_word:
            weight = int(weight * 20)
            temp = []
            for f in string_hash(feature):
                if f == '1':
                    temp.append(weight)
                else:
                    temp.append(-weight)
            key_list.append(temp)
        content_list = np.sum(np.array(key_list), axis=0)
        # 编码读不出来
        if len(key_list) == 0:
            return '00'
        hash_code = ''
        for c in content_list:
            if c > 0:
                hash_code = hash_code + '1'
            else:
                hash_code = hash_code + '0'
        return hash_code

    def _sim_score(self, seq1, seq2):
        """Convert hamming distance to similarity score."""
        # 将距离转化为相似度
        score = 0.0
        if len(seq1) > 2 and len(seq2) > 2:
            score = 1 - hamming_distance(seq1, seq2, normalize=True)
        return score

    def similarity(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """
        Compute hamming similarity between two sentences.

        Parameters
        ----------
        a : str or list of str
        b : str or list of str

        Returns
        -------
        list of float
        """
        if isinstance(a, str):
            a = [a]
        if isinstance(b, str):
            b = [b]
        if len(a) != len(b):
            raise ValueError("expected two inputs of the same length")
        seqs1 = [self.simhash(text) for text in a]
        seqs2 = [self.simhash(text) for text in b]
        scores = [self._sim_score(seq1, seq2) for seq1, seq2 in zip(seqs1, seqs2)]
        return scores

    def distance(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """
        Compute hamming distance between two sentences.

        Parameters
        ----------
        a : str or list of str
        b : str or list of str

        Returns
        -------
        list of float
        """
        sim_scores = self.similarity(a, b)
        return [1 - score for score in sim_scores]

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn: int = 10) -> List[List[Dict]]:
        """
        Find the topn most similar texts to the query against the corpus.
        :param queries: list of str or str
        :param topn: int
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing cosine similarity scores.
        """
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []

        for qid, query in queries.items():
            q_res = []
            query_emb = self.simhash(query)
            for (corpus_id, doc), doc_emb in zip(self.corpus.items(), self.corpus_embeddings):
                score = self._sim_score(query_emb, doc_emb)
                q_res.append({'corpus_id': corpus_id, 'corpus_doc': doc, 'score': score})
            q_res = sorted(q_res, key=lambda x: x['score'], reverse=True)[:topn]
            result.append(q_res)
        return result

    def save_corpus_embeddings(self, emb_path: str = "hash_corpus_emb.jsonl"):
        """
        Save corpus embeddings to jsonl file.
        :param emb_path: jsonl file path
        :return:
        """
        with open(emb_path, "w", encoding="utf-8") as f:
            for id, emb in zip(self.corpus.keys(), self.corpus_embeddings):
                json_obj = {"id": id, "doc": self.corpus[id], "doc_emb": list(emb)}
                f.write(json.dumps(json_obj, ensure_ascii=False) + "\n")
        logger.debug(f"Save corpus embeddings to file: {emb_path}.")

    def load_corpus_embeddings(self, emb_path: str = "hash_corpus_emb.jsonl"):
        """
        Load corpus embeddings from jsonl file.
        :param emb_path: jsonl file path
        :return:
        """
        try:
            with open(emb_path, "r", encoding="utf-8") as f:
                corpus_embeddings = []
                for line in f:
                    json_obj = json.loads(line)
                    self.corpus[int(json_obj["id"])] = json_obj["doc"]
                    corpus_embeddings.append(json_obj["doc_emb"])
                self.corpus_embeddings = corpus_embeddings
        except (IOError, json.JSONDecodeError):
            logger.error("Error: Could not load corpus embeddings from file.")


class TfidfSimilarity(SimilarityABC):
    """
    Compute TFIDF similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    """

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None):
        super().__init__()
        self.corpus = {}

        self.corpus_embeddings = []
        self.tfidf = TFIDF()
        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: Tfidf"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : list of str
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)
        corpus_texts = list(corpus_new.values())
        corpus_embeddings = []
        for sentence in tqdm(corpus_texts, desc="Computing corpus TFIDF"):
            corpus_embeddings.append(self.tfidf.get_tfidf(sentence))
        if self.corpus_embeddings:
            self.corpus_embeddings += corpus_embeddings
        else:
            self.corpus_embeddings = corpus_embeddings
        logger.debug(f"Add {len(corpus_new)} docs, total: {len(self.corpus)}, emb size: {len(self.corpus_embeddings)}")

    def similarity(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """
        Compute cosine similarity score between two sentences.
        :param a:
        :param b:
        :return:
        """
        if isinstance(a, str):
            a = [a]
        if isinstance(b, str):
            b = [b]
        features1 = [self.tfidf.get_tfidf(text) for text in a]
        features2 = [self.tfidf.get_tfidf(text) for text in b]
        return cos_sim(np.array(features1), np.array(features2))

    def distance(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """Compute cosine distance between two keys."""
        return 1 - self.similarity(a, b)

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn: int = 10) -> List[List[Dict]]:
        """Find the topn most similar texts to the query against the corpus.
        :param queries: list of str or str
        :param topn: int
        :return List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing cosine similarity scores.
        """
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []
        queries_texts = list(queries.values())

        queries_embeddings = np.array([self.tfidf.get_tfidf(query) for query in queries_texts], dtype=np.float32)
        corpus_embeddings = np.array(self.corpus_embeddings, dtype=np.float32)
        all_hits = semantic_search(queries_embeddings, corpus_embeddings, top_k=topn)
        for idx, hits in enumerate(all_hits):
            q_res = []
            for hit in hits[0:topn]:
                q_res.append({"corpus_id": hit['corpus_id'],
                              "corpus_doc": self.corpus.get(hit['corpus_id']),
                              "score": hit['score']})
            result.append(q_res)

        return result

    def save_corpus_embeddings(self, emb_path: str = "tfidf_corpus_emb.jsonl"):
        """
        Save corpus embeddings to jsonl file.
        :param emb_path: jsonl file path
        :return:
        """
        with open(emb_path, "w", encoding="utf-8") as f:
            for id, emb in zip(self.corpus.keys(), self.corpus_embeddings):
                json_obj = {"id": id, "doc": self.corpus[id], "doc_emb": list(emb)}
                f.write(json.dumps(json_obj, ensure_ascii=False) + "\n")
        logger.debug(f"Save corpus embeddings to file: {emb_path}.")

    def load_corpus_embeddings(self, emb_path: str = "tfidf_corpus_emb.jsonl"):
        """
        Load corpus embeddings from jsonl file.
        :param emb_path: jsonl file path
        :return:
        """
        try:
            with open(emb_path, "r", encoding="utf-8") as f:
                corpus_embeddings = []
                for line in f:
                    json_obj = json.loads(line)
                    self.corpus[int(json_obj["id"])] = json_obj["doc"]
                    corpus_embeddings.append(json_obj["doc_emb"])
                self.corpus_embeddings = corpus_embeddings
        except (IOError, json.JSONDecodeError):
            logger.error("Error: Could not load corpus embeddings from file.")


class BM25Similarity(SimilarityABC):
    """
    Compute BM25OKapi similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    """

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None):
        super().__init__()
        self.corpus = {}

        self.bm25 = None
        self.default_stopwords = load_stopwords(default_stopwords_file)
        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: BM25"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : dict
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)
        logger.debug(f"Add corpus done, new docs: {len(corpus_new)}, all corpus size: {len(self.corpus)}")
        self.build_bm25()

    def build_bm25(self):
        """build bm25 model."""
        corpus_texts = list(self.corpus.values())
        corpus_seg = [jieba.lcut(d) for d in corpus_texts]
        corpus_seg = [[w for w in doc if (w.strip().lower() not in self.default_stopwords) and
                       len(w.strip()) > 0] for doc in corpus_seg]
        self.bm25 = BM25Okapi(corpus_seg)

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn=10) -> List[List[Dict]]:
        """
        Find the topn most similar texts to the query against the corpus.
        :param queries: input query
        :param topn: int
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing similarity scores.
        """
        if not self.corpus:
            raise ValueError("corpus is None. Please add_corpus first, eg. `add_corpus(corpus)`")
        if not self.bm25:
            self.build_bm25()
        if isinstance(queries, str):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []
        for qid, query in queries.items():
            tokens = jieba.lcut(query)
            scores = self.bm25.get_scores(tokens)

            q_res = [{'corpus_id': corpus_id, 'corpus_doc': self.corpus.get(corpus_id),
                      'score': score} for corpus_id, score in enumerate(scores)]
            q_res = sorted(q_res, key=lambda x: x['score'], reverse=True)[:topn]
            result.append(q_res)

        return result


class WordEmbeddingSimilarity(SimilarityABC):
    """
    Compute Word2Vec similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    """

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None, model_name_or_path="w2v-light-tencent-chinese"):
        """
        Init WordEmbeddingSimilarity.
        :param model_name_or_path: Word2Vec model name or path to model file.
        :param corpus: list of str
        """
        try:
            from text2vec import Word2Vec
        except ImportError:
            raise ImportError("Please install text2vec first, `pip install text2vec`")
        if isinstance(model_name_or_path, str):
            self.keyedvectors = Word2Vec(model_name_or_path)
        elif hasattr(model_name_or_path, "encode"):
            self.keyedvectors = model_name_or_path
        else:
            raise ValueError("model_name_or_path must be ~text2vec.Word2Vec or Word2Vec model name")
        self.corpus = {}

        self.corpus_embeddings = []
        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: Word2Vec"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : list of str
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)

        corpus_texts = list(corpus_new.values())
        corpus_embeddings = self._get_vector(corpus_texts, show_progress_bar=True).tolist()
        if self.corpus_embeddings:
            self.corpus_embeddings += corpus_embeddings
        else:
            self.corpus_embeddings = corpus_embeddings
        logger.debug(f"Add {len(corpus_new)} docs, total: {len(self.corpus)}, emb size: {len(self.corpus_embeddings)}")

    def _get_vector(self, text, show_progress_bar: bool = False) -> np.ndarray:
        return self.keyedvectors.encode(text, show_progress_bar=show_progress_bar)

    def similarity(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """Compute cosine similarity between two texts."""
        v1 = self._get_vector(a)
        v2 = self._get_vector(b)
        return cos_sim(v1, v2)

    def distance(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """Compute cosine distance between two texts."""
        return 1 - self.similarity(a, b)

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn: int = 10) -> List[List[Dict]]:
        """
        Find the topn most similar texts to the query against the corpus.
        :param queries: list of str or str
        :param topn: int
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing similarity scores.
        """
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []
        queries_texts = list(queries.values())

        queries_embeddings = np.array([self._get_vector(query) for query in queries_texts], dtype=np.float32)
        corpus_embeddings = np.array(self.corpus_embeddings, dtype=np.float32)
        all_hits = semantic_search(queries_embeddings, corpus_embeddings, top_k=topn)
        for idx, hits in enumerate(all_hits):
            q_res = []
            for hit in hits[0:topn]:
                q_res.append({"corpus_id": hit['corpus_id'],
                              "corpus_doc": self.corpus.get(hit['corpus_id']),
                              "score": hit['score']})
            result.append(q_res)

        return result

    def save_corpus_embeddings(self, emb_path: str = "word2vec_corpus_emb.jsonl"):
        """
        Save corpus embeddings to jsonl file.
        :param emb_path: jsonl file path
        :return: None, save corpus embeddings to file
        """
        with open(emb_path, "w", encoding="utf-8") as f:
            for id, emb in zip(self.corpus.keys(), self.corpus_embeddings):
                json_obj = {"id": id, "doc": self.corpus[id], "doc_emb": list(emb)}
                f.write(json.dumps(json_obj, ensure_ascii=False) + "\n")
        logger.debug(f"Save corpus embeddings to file: {emb_path}.")

    def load_corpus_embeddings(self, emb_path: str = "word2vec_corpus_emb.jsonl"):
        """
        Load corpus embeddings from jsonl file.
        :param emb_path: jsonl file path
        :return:
        """
        try:
            with open(emb_path, "r", encoding="utf-8") as f:
                corpus_embeddings = []
                for line in f:
                    json_obj = json.loads(line)
                    self.corpus[int(json_obj["id"])] = json_obj["doc"]
                    corpus_embeddings.append(json_obj["doc_emb"])
                self.corpus_embeddings = corpus_embeddings
        except (IOError, json.JSONDecodeError):
            logger.error("Error: Could not load corpus embeddings from file.")


class CilinSimilarity(SimilarityABC):
    """
    Compute Cilin similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    """
    default_cilin_path = os.path.join(pwd_path, 'data/cilin.txt')

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None, cilin_path: str = default_cilin_path):
        super().__init__()
        self.cilin_dict = self.load_cilin_dict(cilin_path)  # Cilin(词林) semantic dictionary
        self.corpus = {}

        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: Cilin"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : list of str
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)
        logger.debug(f"Add {len(corpus_new)} docs, total: {len(self.corpus)}")

    @staticmethod
    def load_cilin_dict(path):
        """加载词林语义词典"""
        sem_dict = {}
        for line in open(path, 'r', encoding='utf-8'):
            line = line.strip()
            terms = line.split(' ')
            sem_type = terms[0]
            words = terms[1:]
            for word in words:
                if word not in sem_dict:
                    sem_dict[word] = sem_type
                else:
                    sem_dict[word] += ';' + sem_type

        for word, sem_type in sem_dict.items():
            sem_dict[word] = sem_type.split(';')
        return sem_dict

    def _word_sim(self, word1, word2):
        """
        比较计算词语之间的相似度，取max最大值
        :param word1:
        :param word2:
        :return:
        """
        sems_word1 = self.cilin_dict.get(word1, [])
        sems_word2 = self.cilin_dict.get(word2, [])
        score_list = [self._semantic_sim(sem_word1, sem_word2) for sem_word1 in sems_word1 for sem_word2 in sems_word2]
        if score_list:
            return max(score_list)
        else:
            return 0

    def _semantic_sim(self, sem1, sem2):
        """
        基于语义计算词语相似度
        :param sem1:
        :param sem2:
        :return:
        """
        sem1 = [sem1[0], sem1[1], sem1[2:4], sem1[4], sem1[5:7], sem1[-1]]
        sem2 = [sem2[0], sem2[1], sem2[2:4], sem2[4], sem2[5:7], sem2[-1]]
        score = 0
        for index in range(len(sem1)):
            if sem1[index] == sem2[index]:
                if index in [0, 1]:
                    score += 3
                elif index == 2:
                    score += 2
                elif index in [3, 4]:
                    score += 1
        return score / 10

    def similarity(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """
        Compute Cilin similarity between two texts.
        :param a:
        :param b:
        :return:
        """
        if isinstance(a, str):
            a = [a]
        if isinstance(b, str):
            b = [b]
        if len(a) != len(b):
            raise ValueError("expected two inputs of the same length")

        def calc_pair_sim(sentence1, sentence2):
            words1 = [word.word for word in jieba.posseg.cut(sentence1) if word.flag[0] not in ['u', 'x', 'w']]
            words2 = [word.word for word in jieba.posseg.cut(sentence2) if word.flag[0] not in ['u', 'x', 'w']]
            score_words1 = []
            score_words2 = []
            for word1 in words1:
                score = max(self._word_sim(word1, word2) for word2 in words2)
                score_words1.append(score)
            for word2 in words2:
                score = max(self._word_sim(word2, word1) for word1 in words1)
                score_words2.append(score)
            similarity_score = max(sum(score_words1) / len(words1), sum(score_words2) / len(words2))
            return similarity_score

        return [calc_pair_sim(sentence1, sentence2) for sentence1, sentence2 in zip(a, b)]

    def distance(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """Compute cosine distance between two texts."""
        return [1 - s for s in self.similarity(a, b)]

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn: int = 10) -> List[List[Dict]]:
        """Find the topn most similar texts to the query against the corpus.
        :param queries: list of str or str
        :param topn: int
        :return: List[List[Dict]], A list with one entry for each query. Each entry is a list of
            dict with the keys 'corpus_id', 'corpus_doc' and 'score', sorted by decreasing cilin similarity scores.
        """
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []
        for qid, query in queries.items():
            q_res = []
            for corpus_id, doc in self.corpus.items():
                score = self.similarity(query, doc)[0]
                q_res.append({'corpus_id': corpus_id, 'corpus_doc': doc, 'score': score})
            q_res = sorted(q_res, key=lambda x: x['score'], reverse=True)[:topn]
            result.append(q_res)
        return result


class HownetSimilarity(SimilarityABC):
    """
    Compute Hownet similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    """
    default_hownet_path = os.path.join(pwd_path, 'data/hownet.txt')

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None, hownet_path: str = default_hownet_path):
        self.hownet_dict = self.load_hownet_dict(hownet_path)  # semantic dictionary
        self.corpus = {}

        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: Hownet"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : list of str
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)
        logger.debug(f"Add {len(corpus_new)} docs, total: {len(self.corpus)}")

    @staticmethod
    def load_hownet_dict(path):
        """加载Hownet语义词典"""
        hownet_dict = {}
        for line in open(path, 'r', encoding='utf-8'):
            words = [word for word in line.strip().replace(' ', '>').replace('\t', '>').split('>') if word != '']
            word = words[0]
            word_def = words[2]
            hownet_dict[word] = word_def.split(',')
        return hownet_dict

    def _semantic_sim(self, sem1, sem2):
        """计算语义相似度"""
        sem_inter = set(sem1).intersection(set(sem2))
        sem_union = set(sem1).union(set(sem2))
        return float(len(sem_inter)) / float(len(sem_union))

    def _word_sim(self, word1, word2):
        """比较两个词语之间的相似度"""
        sems_word1 = self.hownet_dict.get(word1, [])
        sems_words = self.hownet_dict.get(word2, [])
        scores = [self._semantic_sim(sem_word1, sem_word2) for sem_word1 in sems_word1 for sem_word2 in sems_words]
        if scores:
            return max(scores)
        else:
            return 0

    def similarity(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """
        Computer Hownet similarity between two texts.
        :param a:
        :param b:
        :return:
        """
        if isinstance(a, str):
            a = [a]
        if isinstance(b, str):
            b = [b]
        if len(a) != len(b):
            raise ValueError("expected two inputs of the same length")

        def calc_pair_sim(sentence1, sentence2):
            words1 = [word.word for word in jieba.posseg.cut(sentence1) if word.flag[0] not in ['u', 'x', 'w']]
            words2 = [word.word for word in jieba.posseg.cut(sentence2) if word.flag[0] not in ['u', 'x', 'w']]
            score_words1 = []
            score_words2 = []
            for word1 in words1:
                score = max(self._word_sim(word1, word2) for word2 in words2)
                score_words1.append(score)
            for word2 in words2:
                score = max(self._word_sim(word2, word1) for word1 in words1)
                score_words2.append(score)
            similarity_score = max(sum(score_words1) / len(words1), sum(score_words2) / len(words2))
            return similarity_score

        return [calc_pair_sim(sentence1, sentence2) for sentence1, sentence2 in zip(a, b)]

    def distance(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """Compute Hownet distance between two keys."""
        return [1 - s for s in self.similarity(a, b)]

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn: int = 10) -> List[List[Dict]]:
        """Find the topn most similar texts to the query against the corpus."""
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []

        for qid, query in queries.items():
            q_res = []
            for corpus_id, doc in self.corpus.items():
                score = self.similarity(query, doc)[0]
                q_res.append({'corpus_id': corpus_id, 'corpus_doc': doc, 'score': score})

            q_res = sorted(q_res, key=lambda x: x['score'], reverse=True)[:topn]
            result.append(q_res)
        return result


class SameCharsSimilarity(SimilarityABC):
    """
    Compute text chars similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    不考虑文本字符位置顺序，基于相同字符数占比计算相似度
    """

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None):
        super().__init__()
        self.corpus = {}

        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: SameChars"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : list of str
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)
        logger.debug(f"Add {len(corpus_new)} docs, total: {len(self.corpus)}")

    def similarity(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """
        Compute Chars similarity between two texts.
        :param a:
        :param b:
        :return:
        """
        if isinstance(a, str):
            a = [a]
        if isinstance(b, str):
            b = [b]
        if len(a) != len(b):
            raise ValueError("expected two inputs of the same length")

        def calc_pair_sim(sentence1, sentence2):
            if not sentence1 or not sentence2:
                return 0.0
            same = set(sentence1) & set(sentence2)
            similarity_score = max(len(same) / len(set(sentence1)), len(same) / len(set(sentence2)))
            return similarity_score

        return [calc_pair_sim(sentence1, sentence2) for sentence1, sentence2 in zip(a, b)]

    def distance(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """Compute cosine distance between two texts."""
        return [1 - s for s in self.similarity(a, b)]

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn: int = 10) -> List[List[Dict]]:
        """Find the topn most similar texts to the query against the corpus."""
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []
        for qid, query in queries.items():
            q_res = []
            for corpus_id, doc in self.corpus.items():
                score = self.similarity(query, doc)[0]
                q_res.append({'corpus_id': corpus_id, 'corpus_doc': doc, 'score': score})
            q_res = sorted(q_res, key=lambda x: x['score'], reverse=True)[:topn]
            result.append(q_res)
        return result


class SequenceMatcherSimilarity(SimilarityABC):
    """
    Compute text sequence matcher similarity between two sentences and retrieves most
    similar sentence for a given corpus.
    考虑文本字符位置顺序，基于最长公共子串占比计算相似度
    """

    def __init__(self, corpus: Union[List[str], Dict[int, str]] = None):
        super().__init__()
        self.corpus = {}

        if corpus is not None:
            self.add_corpus(corpus)

    def __len__(self):
        """Get length of corpus."""
        return len(self.corpus)

    def __str__(self):
        base = f"Similarity: {self.__class__.__name__}, matching_model: SequenceMatcher"
        if self.corpus:
            base += f", corpus size: {len(self.corpus)}"
        return base

    def add_corpus(self, corpus: Union[List[str], Dict[int, str]]):
        """
        Extend the corpus with new documents.

        Parameters
        ----------
        corpus : list of str
        """
        corpus_new = {}
        start_id = len(self.corpus) if self.corpus else 0
        current_id = start_id  # keep track of the current id
        if isinstance(corpus, list):
            for id, doc in enumerate(corpus):
                if doc not in self.corpus.values():
                    corpus_new[current_id] = doc
                    current_id += 1
        else:
            for id, doc in corpus.items():
                if doc not in self.corpus.values():
                    corpus_new[id] = doc
                    current_id += 1
        if not corpus_new:
            return
        self.corpus.update(corpus_new)
        logger.debug(f"Add {len(corpus_new)} docs, total: {len(self.corpus)}")

    def similarity(self, a: Union[str, List[str]], b: Union[str, List[str]],
                   min_same_len: int = 70, min_same_len_score: float = 0.9):
        """
        Compute Chars similarity between two texts.
        :param a:
        :param b:
        :param min_same_len:
        :param min_same_len_score:
        :return:
        """
        if isinstance(a, str):
            a = [a]
        if isinstance(b, str):
            b = [b]
        if len(a) != len(b):
            raise ValueError("expected two inputs of the same length")

        def calc_pair_sim(sentence1, sentence2):
            if not sentence1 or not sentence2:
                return 0.0
            same_size = longest_match_size(sentence1, sentence2)
            same_score = min_same_len_score if same_size > min_same_len else 0.0
            similarity_score = max(same_size / len(sentence1), same_size / len(sentence2), same_score)
            return similarity_score

        return [calc_pair_sim(sentence1, sentence2) for sentence1, sentence2 in zip(a, b)]

    def distance(self, a: Union[str, List[str]], b: Union[str, List[str]]):
        """Compute cosine distance between two texts."""
        return [1 - s for s in self.similarity(a, b)]

    def most_similar(self, queries: Union[str, List[str], Dict[int, str]], topn: int = 10) -> List[List[Dict]]:
        """Find the topn most similar texts to the query against the corpus."""
        if isinstance(queries, str) or not hasattr(queries, '__len__'):
            queries = [queries]
        if isinstance(queries, list):
            queries = {id: query for id, query in enumerate(queries)}
        result = []
        for qid, query in queries.items():
            q_res = []
            for corpus_id, doc in self.corpus.items():
                score = self.similarity(query, doc)[0]
                q_res.append({'corpus_id': corpus_id, 'corpus_doc': doc, 'score': score})
            q_res = sorted(q_res, key=lambda x: x['score'], reverse=True)[:topn]
            result.append(q_res)
        return result
